#  This code is a part of X-ray: Generate and Analyse (XGA), a module designed for the XMM Cluster Survey (XCS).
#  Last modified by David J Turner (turne540@msu.edu) 24/07/2024, 16:16. Copyright (c) The Contributors

import os
from random import randint
from typing import List, Union

import astropy.units as u
from astropy.units import Quantity

from .run import xspec_call
from .. import OUTPUT, NUM_CORES, COUNTRATE_CONV_SCRIPT
from ..exceptions import NoProductAvailableError, ModelNotAssociatedError, ParameterNotAssociatedError
from ..products import Spectrum
from ..samples.extended import ClusterSample
from ..sas import evselect_spectrum
from ..sas._common import region_setup
from ..sources import BaseSource, GalaxyCluster
from ..utils import ABUND_TABLES


@xspec_call
def cluster_cr_conv(sources: Union[GalaxyCluster, ClusterSample], outer_radius: Union[str, Quantity],
                    inner_radius: Union[str, Quantity] = Quantity(0, 'arcsec'), sim_temp: Quantity = Quantity(3, 'keV'),
                    sim_met: Union[float, List] = 0.3, conv_en: Quantity = Quantity([[0.5, 2.0]], "keV"),
                    abund_table: str = "angr", group_spec: bool = True, min_counts: int = 5, min_sn: float = None,
                    over_sample: float = None, one_rmf: bool = True, num_cores: int = NUM_CORES):
    """
    This function uses the xspec fakeit tool to calculate conversion factors between count rate and
    luminosity for ARFs and RMFs associated with spectra in the given sources. Once complete the conversion
    factors are stored within the relevant XGA spectrum object. If the requested spectra do not already
    exist then they will automatically be generated for you. Please be aware that this function does not
    support calculating conversion factors from AnnularSpectra.

    :param GalaxyCluster sources: The GalaxyCluster objects to calculate conversion factors for.
    :param str/Quantity outer_radius: The name or value of the outer radius of the region that the
        desired spectrum covers (for instance 'r200' would be acceptable for a GalaxyCluster,
        or Quantity(1000, 'kpc')). If 'region' is chosen (to use the regions in region files), then any
        inner radius will be ignored. If you are generating factors for multiple sources then you can
        also pass a Quantity with one entry per source.
    :param str/Quantity inner_radius: The name or value of the outer radius of the region that the
        desired spectrum covers (for instance 'r200' would be acceptable for a GalaxyCluster,
        or Quantity(1000, 'kpc')). If 'region' is chosen (to use the regions in region files), then any
        inner radius will be ignored. By default this is zero arcseconds, resulting in a circular spectrum. If
        you are generating factors for multiple sources then you can also pass a Quantity with one entry per source.
    :param Quantity sim_temp: The temperature(s) to use for the apec model.
    :param float/List sim_met: The metallicity(s) (in solar met) to use for the apec model.
    :param Quantity conv_en: The energy limit pairs to calculate conversion factors for.
    :param str abund_table: The name of the XSPEC abundance table to use.
    :param bool group_spec: A boolean flag that sets whether generated spectra are grouped or not.
    :param float min_counts: If generating a grouped spectrum, this is the minimum number of counts per channel.
        To disable minimum counts set this parameter to None.
    :param float min_sn: If generating a grouped spectrum, this is the minimum signal to noise in each channel.
        To disable minimum signal to noise set this parameter to None.
    :param float over_sample: The minimum energy resolution for each group, set to None to disable. e.g. if
        over_sample=3 then the minimum width of a group is 1/3 of the resolution FWHM at that energy.
    :param bool one_rmf: This flag tells the method whether it should only generate one RMF for a particular
        ObsID-instrument combination - this is much faster in some circumstances, however the RMF does depend
        slightly on position on the detector.
    :param int num_cores: The number of cores to use (if running locally), default is set to 90% of available.
    """
    # This function supports passing both individual sources and sets of sources
    if isinstance(sources, BaseSource):
        sources = [sources]

    if abund_table not in ABUND_TABLES:
        ab_list = ", ".join(ABUND_TABLES)
        raise ValueError("{0} is not in the accepted list of abundance tables; {1}".format(abund_table, ab_list))

    # This just makes sure that spectra matching what has been requested have actually been generated
    evselect_spectrum(sources, outer_radius, inner_radius, group_spec, min_counts, min_sn, over_sample, one_rmf,
                      num_cores)

    # And use the evselect region preparation function to parse the radii properly, and generate
    #  some more predictable radii quantities
    inn_rad_vals, out_rad_vals = region_setup(sources, outer_radius, inner_radius, True, '')[1:]

    # Checks that the luminosity energy bands are pairs of values
    if conv_en.shape[1] != 2:
        raise ValueError("Luminosity energy bands should be supplied in pairs, defined "
                         "like Quantity([[0.5, 2.0], [2.0, 10.0]], 'keV')")
    # Are the lower limits smaller than the upper limits? - Obviously they should be so I check
    elif not all([conv_en[pair_ind, 0] < conv_en[pair_ind, 1] for pair_ind in range(0, conv_en.shape[0])]):
        raise ValueError("Luminosity energy band first entries must be smaller than second entries.")

    # Check that the correct number of temperatures are supplied
    if not sim_temp.isscalar and len(sim_temp) != len(sources):
        raise ValueError("The sim_temp variable must either be scalar or have the "
                         "same number of entries as there are sources.")
    elif not isinstance(sim_met, float) and len(sim_met) != len(sources):
        raise ValueError("The sim_met variable must either be a float or have the "
                         "same number of entries as there are sources.")

    # Hard coding the model currently, tbabs*apec is a good simple descriptor of a cluster
    model = "tbabs*apec"
    # These are the parameter names for this model
    par_names = "{nH kT Abundanc Redshift norm}"
    convert_low_lims = "{" + " ".join(conv_en[:, 0].to("keV").value.astype(str)) + "}"
    convert_upp_lims = "{" + " ".join(conv_en[:, 1].to("keV").value.astype(str)) + "}"

    script_paths = []
    outfile_paths = []
    src_inds = []
    # This function supports passing multiple sources, so we have to setup a script for all of them.
    for s_ind, source in enumerate(sources):
        # This function can take a single temperature to simulate at, or a list of them (one for each source).
        if sim_temp.isscalar:
            the_temp = sim_temp
        else:
            the_temp = sim_temp[s_ind]
        # Equivalent of above but for metallicities
        if isinstance(sim_met, float):
            the_met = sim_met
        else:
            the_met = sim_met[s_ind]

        total_obs_inst = source.num_pn_obs + source.num_mos1_obs + source.num_mos2_obs
        # Find matching spectrum objects associated with the current source, and checking if they are valid
        spec_objs = source.get_spectra(out_rad_vals[s_ind], inner_radius=inn_rad_vals[s_ind],
                                       group_spec=group_spec, min_counts=min_counts, min_sn=min_sn,
                                       over_sample=over_sample)

        # This is because many other parts of this function assume that spec_objs is iterable, and in the case of
        #  a source with only a single valid instrument for a single valid observation this may not be the case
        if isinstance(spec_objs, Spectrum):
            spec_objs = [spec_objs]

        # Obviously we can't do a fit if there are no spectra, so throw an error if that's the case
        if len(spec_objs) == 0:
            raise NoProductAvailableError("There are no matching spectra for {} object, you "
                                          "need to generate them first!".format(source.name))
        elif len(spec_objs) != total_obs_inst:
            raise NoProductAvailableError("The number of matching spectra ({0}) is not equal to the number of "
                                          "instrument/observation combinations ({1}) for {2}.".format(len(spec_objs),
                                                                                                      total_obs_inst,
                                                                                                      source.name))

        # Turn RMF and ARF paths into TCL style list for substitution into template
        rmf_paths = "{" + " ".join([spec.rmf for spec in spec_objs]) + "}"
        arf_paths = "{" + " ".join([spec.arf for spec in spec_objs]) + "}"
        # Put in the ObsIDs and Instruments, to help name columns easily
        obs = "{" + " ".join([spec.obs_id for spec in spec_objs]) + "}"
        inst = "{" + " ".join([spec.instrument for spec in spec_objs]) + "}"

        # For this model, we have to know the redshift of the source.
        if source.redshift is None:
            raise ValueError("You cannot supply a source without a redshift to this model.")

        t = the_temp.to("keV", equivalencies=u.temperature_energy()).value
        # Another TCL list, this time of the parameter start values for this model.
        par_values = "{{{0} {1} {2} {3} {4}}}".format(source.nH.to("10^22 cm^-2").value, t,
                                                      sim_met, source.redshift, 1.)

        with open(COUNTRATE_CONV_SCRIPT, 'r') as c_script:
            script = c_script.read()

        dest_dir = OUTPUT + "XSPEC/" + source.name + "/"
        if not os.path.exists(dest_dir):
            os.makedirs(dest_dir)
        out_file = dest_dir + source.name + "_" + spec_objs[0].storage_key + "_" + model + "_conv_factors.csv"
        script_file = dest_dir + source.name + "_" + spec_objs[0].storage_key + "_" + model + "_conv_factors" + ".xcm"

        # Random ident to make sure no temporary spec files clash
        rid = randint(0, int(1e+8))

        # Populates the fakeit conversion factor template script
        script = script.format(ab=abund_table, H0=source.cosmo.H0.value, q0=0., lamb0=source.cosmo.Ode0,
                               rmf=rmf_paths, arf=arf_paths, obs=obs, inst=inst, m=model, pn=par_names,
                               pv=par_values, lll=convert_low_lims, lul=convert_upp_lims,
                               redshift=source.redshift, of=out_file, rid=rid)

        # Write out the filled-in template to its destination
        with open(script_file, 'w') as xcm:
            xcm.write(script)

        try:
            # Checks through the spectrum objects we retrieved earlier, and the energy limits,
            #  to look for conversion factor results, if they exist they aren't run again, otherwise an error
            #  is triggered and the scripts get added to the pile to run.
            res = [s.get_conv_factor(e_pair[0], e_pair[1], "tbabs*apec") for e_pair in conv_en for s in spec_objs]
        except (ModelNotAssociatedError, ParameterNotAssociatedError):
            script_paths.append(script_file)
            outfile_paths.append(out_file)
            src_inds.append(s_ind)

    # New feature of XSPEC interface, tells the xspec_call decorator what type of output from the script
    #  to expect
    run_type = "conv_factors"
    # I don't allow the user to set a timeout for fakeit runs because its unnecessary, but the run code still
    #  needs a value
    return script_paths, outfile_paths, num_cores, run_type, src_inds, None, Quantity(1, 'hour')









#  This code is a part of X-ray: Generate and Analyse (XGA), a module designed for the XMM Cluster Survey (XCS).
#  Last modified by David J Turner (turne540@msu.edu) 09/08/2024, 13:44. Copyright (c) The Contributors

from typing import Union, List, Tuple
from warnings import warn

import numpy as np
from abel.basex import basex_transform
from abel.dasch import onion_peeling_transform, two_point_transform, three_point_transform
from abel.daun import daun_transform
from abel.direct import direct_transform
from abel.hansenlaw import hansenlaw_transform
from abel.onion_bordas import onion_bordas_transform
from astropy.constants import m_p
from astropy.units import Quantity, kpc
from tqdm import tqdm

from .misc import model_check
from .temperature import min_snr_proj_temp_prof, min_cnt_proj_temp_prof, ALLOWED_ANN_METHODS
from ..exceptions import NoProductAvailableError, ModelNotAssociatedError, ParameterNotAssociatedError
from ..imagetools.profile import radial_brightness
from ..models import BaseModel1D
from ..products.profile import SurfaceBrightness1D, GasDensity3D
from ..samples.extended import ClusterSample
from ..sas._common import region_setup
from ..sources import GalaxyCluster, BaseSource
from ..sourcetools import ang_to_rad
from ..utils import NHC, ABUND_TABLES, NUM_CORES, MEAN_MOL_WEIGHT
from ..xspec.fakeit import cluster_cr_conv
from ..xspec.fit import single_temp_apec

ALLOWED_INV_ABEL = ['direct', 'basex', 'hansen_law_ho0', 'hansen_law_ho1', 'onion_bordas', 'onion_peeling',
                    'two_point', 'three_point', 'daun']


def _dens_setup(sources: Union[GalaxyCluster, ClusterSample], outer_radius: Union[str, Quantity],
                inner_radius: Union[str, Quantity], abund_table: str, lo_en: Quantity,
                hi_en: Quantity, group_spec: bool = True, min_counts: int = 5, min_sn: float = None,
                over_sample: float = None, obs_id: Union[str, list] = None, inst: Union[str, list] = None,
                conv_temp: Quantity = None, conv_outer_radius: Quantity = "r500",
                num_cores: int = NUM_CORES) -> Tuple[Union[ClusterSample, List], List[Quantity], list, list]:
    """
    An internal function which exists because all the density profile methods that I have planned
    need the same product checking and setup steps. This function checks that all necessary spectra/fits have
    been generated/run, then uses them to calculate the conversion factors from count-rate/volume to squared
    hydrogen number density.

    :param Union[GalaxyCluster, ClusterSample] sources: The source objects/sample object for which the density profile
    is being found.
    :param str/Quantity outer_radius: The name or value of the outer radius of the spectra that should be used
        to calculate conversion factors (for instance 'r200' would be acceptable for a GalaxyCluster, or
        Quantity(1000, 'kpc')). If 'region' is chosen (to use the regions in region files), then any
        inner radius will be ignored.
    :param str/Quantity inner_radius: The name or value of the inner radius of the spectra that should be used
        to calculate conversion factors (for instance 'r500' would be acceptable for a GalaxyCluster, or
        Quantity(300, 'kpc')). By default this is zero arcseconds, resulting in a circular spectrum.
    :param str abund_table: Which abundance table should be used for the XSPEC fit, FakeIt run, and for the
        electron/hydrogen number density ratio.
    :param Quantity lo_en: The lower energy limit of the combined ratemap used to calculate density.
    :param Quantity hi_en: The upper energy limit of the combined ratemap used to calculate density.
    :param bool group_spec: Whether the spectra that were fitted for the desired result were grouped.
    :param float min_counts: The minimum counts per channel, if the spectra that were fitted for the
        desired result were grouped by minimum counts.
    :param float min_sn: The minimum signal-to-noise per channel, if the spectra that were fitted for the
        desired result were grouped by minimum signal-to-noise.
    :param float over_sample: The level of oversampling applied on the spectra that were fitted.
    :param str/list obs_id: A specific ObsID(s) to measure the density from. This should be a string if a single
        source is being analysed, and a list of ObsIDs the same length as the number of sources otherwise. The
        default is None, in which case the combined data will be used to measure the density profile.
    :param str/list inst: A specific instrument(s) to measure the density from. This can either be passed as a
        single string (e.g. 'pn') if only one source is being analysed, or the same instrument should be used for
        every source in a sample, or a list of strings if different instruments are required for each source. The
        default is None, in which case the combined data will be used to measure the density profile.
    :param Quantity conv_temp: If set this will override XGA measured temperatures within the conv_outer_radius, and
        the fakeit run to calculate the normalisation conversion factor will use these temperatures. The quantity
         should have an entry for each cluster being analysed. Default is None.
    :param str/Quantity conv_outer_radius: The outer radius within which to generate spectra and measure temperatures
        for the conversion factor calculation, default is 'r500'. An astropy quantity may also be passed, with either
        a single value or an entry for each cluster being analysed.
    :param int num_cores: The number of cores that the evselect call and XSPEC functions are allowed to use.
    :return: The source object(s)/sample that was passed in, an array of the calculated conversion factors to take the
        count-rate/volume to a number density of hydrogen, the parsed obs_id variable, and the parsed inst variable.
    :rtype: Tuple[Union[ClusterSample, List], List[Quantity], list, list]
    """
    # If its a single source I shove it in a list so I can just iterate over the sources parameter
    #  like I do when its a Sample object
    if isinstance(sources, BaseSource):
        sources = [sources]

    # Perform some checks on the ObsID and instrument parameters to make sure that they are in the correct
    #  format if they have been set. We don't need to check that the ObsIDs are associated with the sources
    #  here, because that will happen when the ratemaps are retrieved from the source objects.
    if all([obs_id is not None, inst is not None]):
        if isinstance(obs_id, str):
            obs_id = [obs_id]
        if isinstance(inst, str):
            inst = [inst]

        if len(obs_id) != len(sources):
            raise ValueError("If you set the obs_id argument there must be one entry per source being analysed.")

        if len(inst) != len(sources) and len(inst) != 1:
            raise ValueError("The value passed for inst must either be a single instrument name, or a list "
                             "of instruments the same length as the number of sources being analysed.")
    elif all([obs_id is None, inst is None]):
        obs_id = [None]*len(sources)
        inst = [None]*len(sources)
    else:
        raise ValueError("If a value is supplied for obs_id, then a value must be supplied for inst as well, and "
                         "vice versa.")

    if not all([type(src) == GalaxyCluster for src in sources]):
        raise TypeError("Only GalaxyCluster sources can be passed to cluster_density_profile.")

    # Triggers an exception if the abundance table name passed isn't recognised
    if abund_table not in ABUND_TABLES:
        ab_list = ", ".join(ABUND_TABLES)
        raise ValueError("{0} is not in the accepted list of abundance tables; {1}".format(abund_table, ab_list))

    # This check will eventually become obselete, but I haven't yet implemented electron to proton ratios for
    #  all allowed abundance tables - so this just checks whether the chosen table has an ratio associated.
    try:
        e_to_p_ratio = NHC[abund_table]
    except KeyError:
        raise NotImplementedError("That is an acceptable abundance table, but I haven't added the conversion factor "
                                  "to the dictionary yet")

    if conv_temp is not None and not conv_temp.isscalar and len(conv_temp) != len(sources):
        raise ValueError("If multiple there are multiple entries in conv_temp, then there must be the same number"
                         " of entries as there are sources being analysed.")
    elif conv_temp is not None:
        temps = conv_temp
    else:
        # Check that the spectra we will be relying on for conversion calculation have been fitted, calling
        #  this function will also make sure that they are generated
        single_temp_apec(sources, conv_outer_radius, inner_radius, abund_table=abund_table, group_spec=group_spec,
                         min_counts=min_counts, min_sn=min_sn, over_sample=over_sample, num_cores=num_cores)

        # Then we need to grab the temperatures and pass them through to the cluster conversion factor
        #  calculator - this may well change as I intend to let cluster_cr_conv grab temperatures for
        #  itself at some point
        temp_temps = []
        for src in sources:
            try:
                # A temporary temperature variable
                temp_temp = src.get_temperature(conv_outer_radius, "constant*tbabs*apec", inner_radius, group_spec,
                                                min_counts, min_sn, over_sample)[0]
            except (ModelNotAssociatedError, ParameterNotAssociatedError):
                warn("{s}'s temperature fit is not valid, so I am defaulting to a temperature of "
                     "3keV".format(s=src.name))
                temp_temp = Quantity(3, 'keV')
            temp_temps.append(temp_temp.value)
        temps = Quantity(temp_temps, 'keV')

    # This call actually does the fakeit calculation of the conversion factors, then stores them in the
    #  XGA Spectrum objects
    cluster_cr_conv(sources, conv_outer_radius, inner_radius, temps, abund_table=abund_table, num_cores=num_cores,
                    group_spec=group_spec, min_counts=min_counts, min_sn=min_sn, over_sample=over_sample)

    # This where the combined conversion factor that takes a count-rate/volume to a squared number density
    #  of hydrogen
    to_dens_convs = []
    # These are from the distance and redshift, also the normalising 10^-14 (see my paper for
    #  more of an explanation)
    for src_ind, src in enumerate(sources):
        src: GalaxyCluster
        # Both the angular_diameter_distance and redshift are guaranteed to be present here because redshift
        #  is REQUIRED to define GalaxyCluster objects
        factor = ((4 * np.pi * (src.angular_diameter_distance.to("cm") * (1 + src.redshift)) ** 2)
                  / (e_to_p_ratio * 10 ** -14))
        total_factor = factor * src.norm_conv_factor(conv_outer_radius, lo_en, hi_en, inner_radius, group_spec,
                                                     min_counts, min_sn, over_sample, obs_id[src_ind], inst[src_ind])
        to_dens_convs.append(total_factor)

    return sources, to_dens_convs, obs_id, inst


def _run_sb(src: GalaxyCluster, outer_radius: Quantity, use_peak: bool, lo_en: Quantity, hi_en: Quantity,
            psf_corr: bool, psf_model: str, psf_bins: int, psf_algo: str, psf_iter: int, pix_step: int,
            min_snr: float, obs_id: str = None, inst: str = None) -> SurfaceBrightness1D:
    """
    An internal function for the Surface Brightness based density functions, which just quickly assembles the
    requested surface brightness profile.

    :param GalaxyCluster src: A GalaxyCluster object to generate a brightness profile for.
    :param Quantity outer_radius: The desired outer radius of the brightness profile.
    :param bool use_peak: If true the measured peak will be used as the central coordinate of the profile.
    :param Quantity lo_en: The lower energy limit of the combined ratemap used to calculate density.
    :param Quantity hi_en: The upper energy limit of the combined ratemap used to calculate density.
    :param bool psf_corr: Default True, whether PSF corrected ratemaps will be used to make the
        surface brightness profile, and thus the density (if False density results could be incorrect).
    :param str psf_model: If PSF corrected, the PSF model used.
    :param int psf_bins: If PSF corrected, the number of bins per side.
    :param str psf_algo: If PSF corrected, the algorithm used.
    :param int psf_iter: If PSF corrected, the number of algorithm iterations.
    :param int pix_step: The width (in pixels) of each annular bin for the profiles, default is 1.
    :param int/float min_snr: The minimum allowed signal-to-noise for the surface brightness
        profiles. Default is 0, which disables automatic re-binning.
    :param str obs_id: The ObsID of the ratemap that should be used to generate the brightness profile, default
        is None in which case the combined ratemap will be used.
    :param str inst: The instrument of the ratemap that should be used to generate the brightness profile, default
        is None in which case the combined ratemap will be used.
    :return: The requested surface brightness profile.
    :rtype: SurfaceBrightness1D
    """
    try:
        if all([obs_id is None, inst is None]):
            rt = src.get_combined_ratemaps(lo_en, hi_en, psf_corr, psf_model, psf_bins, psf_algo, psf_iter)
            # Grabs the mask which will remove interloper sources
            int_mask = src.get_interloper_mask()
            comb = True
        elif all([obs_id is not None, inst is not None]):
            rt = src.get_ratemaps(obs_id, inst, lo_en, hi_en, psf_corr, psf_model, psf_bins, psf_algo, psf_iter)
            # Grabs the mask which will remove interloper sources
            int_mask = src.get_interloper_mask(obs_id=obs_id)
            comb = False
        else:
            raise ValueError("If an ObsID is supplied, an instrument must be supplied as well, and vice versa.")
    except NoProductAvailableError:
        raise NoProductAvailableError("The RateMap required to measure the density profile has not been generated "
                                      "yet, possibly because you haven't generated PSF corrected image yet.")

    if use_peak:
        centre = src.peak
    else:
        centre = src.ra_dec

    rad = src.convert_radius(outer_radius, 'kpc')

    try:
        sb_prof = src.get_1d_brightness_profile(rad, obs_id, inst, centre, lo_en=lo_en, hi_en=hi_en,
                                                pix_step=pix_step, min_snr=min_snr, psf_corr=psf_corr,
                                                psf_model=psf_model, psf_bins=psf_bins, psf_algo=psf_algo,
                                                psf_iter=psf_iter)
    except NoProductAvailableError:
        try:
            sb_prof, success = radial_brightness(rt, centre, rad, src.background_radius_factors[0],
                                                 src.background_radius_factors[1], int_mask, src.redshift, pix_step,
                                                 kpc, src.cosmo, min_snr)
        except ValueError:
            sb_prof = None
            success = False
            # No longer just background region failure that can set this off
            # warn("Background region for brightness profile is all zeros for {}".format(src.name))

        if sb_prof is not None and not success:
            warn("Minimum SNR rebinning failed for {}".format(src.name), stacklevel=2)

    return sb_prof


def inv_abel_fitted_model(sources: Union[GalaxyCluster, ClusterSample],
                          model: Union[str, List[str], BaseModel1D, List[BaseModel1D]], fit_method: str = "mcmc",
                          outer_radius: Union[str, Quantity] = "r500", num_dens: bool = True, use_peak: bool = True,
                          pix_step: int = 1, min_snr: Union[int, float] = 0.0, abund_table: str = "angr",
                          lo_en: Quantity = Quantity(0.5, 'keV'), hi_en: Quantity = Quantity(2.0, 'keV'),
                          psf_corr: bool = True, psf_model: str = "ELLBETA", psf_bins: int = 4, psf_algo: str = "rl",
                          psf_iter: int = 15, num_walkers: int = 20, num_steps: int = 20000, num_samples: int = 10000,
                          group_spec: bool = True, min_counts: int = 5, min_sn: float = None, over_sample: float = None,
                          obs_id: Union[str, list] = None, inst: Union[str, list] = None, conv_temp: Quantity = None,
                          conv_outer_radius: Quantity = "r500", inv_abel_method: str = None, num_cores: int = NUM_CORES,
                          show_warn: bool = True) -> List[GasDensity3D]:
    """
    A count-rate-map-based galaxy cluster gas density calculation method where a surface brightness profile is fit with
    a model and an inverse abel transform is used to infer the 3D count-rate/volume profile. Then a conversion factor
    calculated from simulated spectra is used to infer the number density profile.

    Depending on the chosen surface brightness model, the inverse abel transform may be performed using an analytical
    solution, or numerical methods.

    :param GalaxyCluster/ClusterSample sources: A GalaxyCluster or ClusterSample object to measure density
        profiles for.
    :param str/List[str]/BaseModel1D/List[BaseModel1D] model: The model(s) to be fit to the cluster surface
        profile(s). You may pass the string name of a model (for single or multiple clusters), a single instance
        of an XGA model class (for single or multiple clusters), a list of string names (one entry for each cluster
        being analysed), or a list of XGA model instances (one entry for each cluster being analysed).
    :param str fit_method: The method for the profile object to use to fit the model, default is mcmc.
    :param str/Quantity outer_radius: The radius to which the surface brightness profile should be generated. It can
        be a named radius (e.g. 'r500', 'r200') or a quantity containing a value (or values, if a sample of clusters
        has been passed).
    :param bool num_dens: If True then a number density profile will be generated, otherwise a mass density profile
        will be generated.
    :param bool use_peak: If true the measured peak will be used as the central coordinate of the profile.
    :param int pix_step: The width (in pixels) of each annular bin for the profiles, default is 1.
    :param int/float min_snr: The minimum allowed signal-to-noise for the surface brightness
        profiles. Default is 0, which disables automatic re-binning.
    :param str abund_table: Which abundance table should be used for the XSPEC fit, FakeIt run, and for the
        electron/hydrogen number density ratio.
    :param Quantity lo_en: The lower energy limit of the combined ratemap used to calculate density.
    :param Quantity hi_en: The upper energy limit of the combined ratemap used to calculate density.
    :param bool psf_corr: Default True, whether PSF corrected ratemaps will be used to make the
        surface brightness profile, and thus the density (if False density results could be incorrect).
    :param str psf_model: If PSF corrected, the PSF model used.
    :param int psf_bins: If PSF corrected, the number of bins per side.
    :param str psf_algo: If PSF corrected, the algorithm used.
    :param int psf_iter: If PSF corrected, the number of algorithm iterations.
    :param int num_walkers: If using mcmc fitting, the number of walkers to use. Default is 20.
    :param int num_steps: If using mcmc fitting, the number of steps each walker should take. Default is 20000.
    :param int num_samples: The number of samples drawn from the posterior distributions of model parameters
        after the fitting process is complete.
    :param bool group_spec: Whether the spectra that were used for fakeit were grouped.
    :param float min_counts: The minimum counts per channel, if the spectra that were used for fakeit
        were grouped by minimum counts.
    :param float min_sn: The minimum signal-to-noise per channel, if the spectra that were used for fakeit
        were grouped by minimum signal-to-noise.
    :param float over_sample: The level of oversampling applied on the spectra that were used for fakeit.
    :param str/list obs_id: A specific ObsID(s) to measure the density from. This should be a string if a single
        source is being analysed, and a list of ObsIDs the same length as the number of sources otherwise. The
        default is None, in which case the combined data will be used to measure the density profile.
    :param str/list inst: A specific instrument(s) to measure the density from. This can either be passed as a
        single string (e.g. 'pn') if only one source is being analysed, or the same instrument should be used for
        every source in a sample, or a list of strings if different instruments are required for each source. The
        default is None, in which case the combined data will be used to measure the density profile.
    :param Quantity conv_temp: If set this will override XGA measured temperatures within the conv_outer_radius, and
        the fakeit run to calculate the normalisation conversion factor will use these temperatures. The quantity
         should have an entry for each cluster being analysed. Default is None.
    :param str/Quantity conv_outer_radius: The outer radius within which to generate spectra and measure temperatures
        for the conversion factor calculation, default is 'r500'. An astropy quantity may also be passed, with either
        a single value or an entry for each cluster being analysed.
    :param str inv_abel_method: The method which should be used for the inverse abel transform of model which
        is fitted to the surface brightness profile. This overrides the default method for the model, which is either
        'analytical' for models with an analytical solution to the inverse abel transform, or 'direct' for
        models which don't have an analytical solution. Default is None.
    :param int num_cores: The number of cores that the evselect call and XSPEC functions are allowed to use.
    :param bool show_warn: Should fit warnings be shown on screen.
    :return: A list of the 3D gas density profiles measured by this function, though if the measurement was not
        successful an entry of None will be added to the list.
    :rtype: List[GasDensity3D]
    """
    # Run the setup function, calculates the factors that translate 3D countrate to density
    #  Also checks parameters and runs any spectra/fits that need running
    sources, conv_factors, obs_id, inst = _dens_setup(sources, outer_radius, Quantity(0, 'arcsec'), abund_table, lo_en,
                                                      hi_en, group_spec, min_counts, min_sn, over_sample, obs_id, inst,
                                                      conv_temp, conv_outer_radius, num_cores)

    # Calls the handy spectrum region setup function to make a predictable set of outer radius values
    out_rads = region_setup(sources, outer_radius, Quantity(0, 'arcsec'), False, '')[-1]

    # Need to sort out the type of model input that the user chose, and make sure its ready to be passed into
    #  the fit method of the surface brightness profile(s)
    # First we check the number of arguments passed for the model
    model = model_check(sources, model)

    with tqdm(desc="Fitting data, inverse Abel transforming, and measuring densities",
              total=len(sources), position=0) as dens_prog:
        final_dens_profs = []
        # I need the ratio of electrons to protons here as well, so just fetch that for the current abundance table
        e_to_p_ratio = NHC[abund_table]
        for src_ind, src in enumerate(sources):

            sb_prof = _run_sb(src, out_rads[src_ind], use_peak, lo_en, hi_en, psf_corr, psf_model, psf_bins, psf_algo,
                              psf_iter, pix_step, min_snr, obs_id[src_ind], inst[src_ind])
            if sb_prof is None:
                final_dens_profs.append(None)
                continue
            else:
                src.update_products(sb_prof)

            # Fit the user chosen model to sb_prof
            cur_model = model[src_ind]
            sb_prof.fit(cur_model, fit_method, num_samples, num_steps, num_walkers, show_warn=show_warn,
                        progress_bar=False)

            if isinstance(cur_model, str):
                model_r = sb_prof.get_model_fit(cur_model, fit_method)
            else:
                model_r = sb_prof.get_model_fit(cur_model.name, fit_method)

            if model_r.success:
                dens_rads = sb_prof.radii.copy()
                dens_rads_errs = sb_prof.radii_err.copy()
                dens_deg_rads = sb_prof.deg_radii.copy()
                # Run the inverse abel transform for the model, to retrieve distributions for the value of the transformed
                #  model at each r point. If the user hasn't set a method then we use the default method for the current
                #  model, otherwise we pass the user's choice
                if inv_abel_method is None:
                    transformed = model_r.inverse_abel(dens_rads, use_par_dist=True)
                else:
                    transformed = model_r.inverse_abel(dens_rads, use_par_dist=True, method=inv_abel_method)

                # Now need to make sure the units of the transformed model are what we need
                if sb_prof.values_unit.is_equivalent('ct/(s*arcmin**2)'):
                    # If the SB profile is in count/s/arcmin^2 then the abel transform will have
                    #  units of ct/s/(arcmin^2 kpc), so I create a quantity which will convert the arcmin^2 to kpc^2
                    conv = Quantity(ang_to_rad(Quantity(1, 'arcmin'), src.redshift, src.cosmo).to("kpc").value,
                                    'kpc/arcmin')**2
                    transformed /= conv
                elif sb_prof.values_unit.is_equivalent('ct/(s*kpc**2)'):
                    pass
                else:
                    raise NotImplementedError("Haven't yet added support for surface brightness profiles in "
                                              "other units, don't really know how you even got here.")

                # We convert the volume element to cm^3 now, this is the unit we expect for the density conversion
                transformed = transformed.to('ct/(s*cm^3)')

                # We multiply by the conversion factor that is unique to the cluster and calculated earlier to take
                #  the transformed profile to a gas number density (n_gas as seen in Eckert et al. 2016, eq. 2).
                num_dens_dist = np.sqrt(transformed * conv_factors[src_ind])*(1+e_to_p_ratio)

                med_num_dens = np.percentile(num_dens_dist, 50, axis=1)
                num_dens_err = np.std(num_dens_dist, axis=1)

                # Setting up the instrument and ObsID to pass into the density profile definition
                if obs_id[src_ind] is None:
                    cur_inst = "combined"
                    cur_obs = "combined"
                else:
                    cur_inst = inst[src_ind]
                    cur_obs = obs_id[src_ind]

                try:
                    # I now allow the user to decide if they want to generate number or mass density profiles using
                    #  this function, and here is where that distinction is made
                    if num_dens:
                        dens_prof = GasDensity3D(dens_rads.to("kpc"), med_num_dens, sb_prof.centre, src.name, cur_obs,
                                                 cur_inst, model_r.name, sb_prof, dens_rads_errs, num_dens_err,
                                                 deg_radii=dens_deg_rads, auto_save=True)
                    else:
                        # TODO Check the origin of the mean molecular weight, see if there are different values for
                        #  different abundance tables
                        # The mean molecular weight multiplied by the proton mass
                        conv_mass = MEAN_MOL_WEIGHT*m_p
                        dens_prof = GasDensity3D(dens_rads.to("kpc"), (med_num_dens*conv_mass).to('Msun/Mpc^3'),
                                                 sb_prof.centre, src.name, cur_obs, cur_inst, model_r.name, sb_prof,
                                                 dens_rads_errs, (num_dens_err*conv_mass).to('Msun/Mpc^3'),
                                                 deg_radii=dens_deg_rads, auto_save=True)

                    src.update_products(dens_prof)
                    final_dens_profs.append(dens_prof)

                # If, for some reason, there are some inf/NaN values in any of the quantities passed to the GasDensity3D
                #  declaration, this is where an error will be thrown
                except ValueError:
                    final_dens_profs.append(None)
                    warn("One or more of the quantities passed to the init of {}'s density profile has a NaN or Inf "
                         "value in it.".format(src.name), stacklevel=2)
            else:
                final_dens_profs.append(None)

            dens_prog.update(1)

    return final_dens_profs


def ann_spectra_apec_norm(sources: Union[GalaxyCluster, ClusterSample], outer_radii: Union[Quantity, List[Quantity]],
                          num_dens: bool = True, annulus_method: str = 'min_snr', min_snr: float = 30,
                          min_cnt: Union[int, Quantity] = Quantity(1000, 'ct'),
                          min_width: Quantity = Quantity(20, 'arcsec'), use_combined: bool = True,
                          use_worst: bool = False, lo_en: Quantity = Quantity(0.5, 'keV'),
                          hi_en: Quantity = Quantity(2, 'keV'), psf_corr: bool = False, psf_model: str = "ELLBETA",
                          psf_bins: int = 4, psf_algo: str = "rl", psf_iter: int = 15, allow_negative: bool = False,
                          exp_corr: bool = True, group_spec: bool = True, min_counts: int = 5, min_sn: float = None,
                          over_sample: float = None, one_rmf: bool = True, freeze_met: bool = True,
                          abund_table: str = "angr", temp_lo_en: Quantity = Quantity(0.3, 'keV'),
                          temp_hi_en: Quantity = Quantity(7.9, 'keV'), num_data_real: int = 10000, sigma: int = 1,
                          num_cores: int = NUM_CORES) -> List[GasDensity3D]:
    """
    A method of measuring density profiles using XSPEC fits of a set of Annular Spectra. First checks whether the
    required annular spectra already exist and have been fit using XSPEC, if not then they are generated and fitted,
    and APEC normalisation profiles will be produced (with projected temperature profiles also being made as a useful
    extra). Then the apec normalisation profile will be used, with knowledge of the source's redshift and chosen
    analysis cosmology, to produce a density profile from the APEC normalisation.

    :param GalaxyCluster/ClusterSample sources: An individual or sample of sources to calculate 3D gas
        density profiles for.
    :param str/Quantity outer_radii: The name or value of the outer radius to use for the generation of
        the spectrum (for instance 'r200' would be acceptable for a GalaxyCluster, or Quantity(1000, 'kpc')). If
        'region' is chosen (to use the regions in region files), then any inner radius will be ignored. If you are
        generating for multiple sources then you can also pass a Quantity with one entry per source.
    :param bool num_dens: If True then a number density profile will be generated, otherwise a mass density profile
        will be generated.
    :param str annulus_method: The method by which the annuli are designated, this can be 'min_snr' (which will use
        the min_snr_proj_temp_prof function), or 'min_cnt' (which will use the min_cnt_proj_temp_prof function).
    :param float min_snr: The minimum signal-to-noise which is allowable in a given annulus, used if annulus_method
        is set to 'min_snr'.
    :param int/Quantity min_cnt: The minimum background subtracted counts which are allowable in a given annulus, used
        if annulus_method is set to 'min_cnt'.
    :param Quantity min_width: The minimum allowable width of an annulus. The default is set to 20 arcseconds to try
        and avoid PSF effects.
    :param bool use_combined: If True (and annulus_method is set to 'min_snr') then the combined RateMap will be
        used for signal-to-noise annulus calculations, this is overridden by use_worst. If True (and annulus_method
        is set to 'min_cnt') then combined RateMaps will be used for annulus count calculations, if False then
        the median observation (in terms of counts) will be used.
    :param bool use_worst: If True then the worst observation of the cluster (ranked by global signal-to-noise) will
        be used for signal-to-noise annulus calculations. Used if annulus_method is set to 'min_snr'.
    :param Quantity lo_en: The lower energy bound of the RateMap to use for the signal-to-noise or background
        subtracted count calculations.
    :param Quantity hi_en: The upper energy bound of the RateMap to use for the signal-to-noise or background
        subtracted count calculations.
    :param bool psf_corr: Sets whether you wish to use a PSF corrected RateMap or not.
    :param str psf_model: If the RateMap you want to use is PSF corrected, this is the PSF model used.
    :param int psf_bins: If the RateMap you want to use is PSF corrected, this is the number of PSFs per
        side in the PSF grid.
    :param str psf_algo: If the RateMap you want to use is PSF corrected, this is the algorithm used.
    :param int psf_iter: If the RateMap you want to use is PSF corrected, this is the number of iterations.
    :param bool allow_negative: Should pixels in the background subtracted count map be allowed to go below
        zero, which results in a lower signal-to-noise (and can result in a negative signal-to-noise).
    :param bool exp_corr: Should signal-to-noises be measured with exposure time correction, default is True. I
            recommend that this be true for combined observations, as exposure time could change quite dramatically
            across the combined product.
    :param bool group_spec: A boolean flag that sets whether generated spectra are grouped or not.
    :param float min_counts: If generating a grouped spectrum, this is the minimum number of counts per channel.
        To disable minimum counts set this parameter to None.
    :param float min_sn: If generating a grouped spectrum, this is the minimum signal-to-noise in each channel.
        To disable minimum signal-to-noise set this parameter to None.
    :param float over_sample: The minimum energy resolution for each group, set to None to disable. e.g. if
        over_sample=3 then the minimum width of a group is 1/3 of the resolution FWHM at that energy.
    :param bool one_rmf: This flag tells the method whether it should only generate one RMF for a particular
        ObsID-instrument combination - this is much faster in some circumstances, however the RMF does depend
        slightly on position on the detector.
    :param bool freeze_met: Whether the metallicity parameter in the fits to annuli in XSPEC should be frozen.
    :param str abund_table: The abundance table to use both for the conversion from n_exn_p to n_e^2 during density
        calculation, and the XSPEC fit.
    :param Quantity temp_lo_en: The lower energy limit for the XSPEC fits to annular spectra.
    :param Quantity temp_hi_en: The upper energy limit for the XSPEC fits to annular spectra.
    :param int num_data_real: The number of random realisations to generate when propagating profile uncertainties.
    :param int sigma: What sigma uncertainties should newly created profiles have, the default is 2σ.
    :param int num_cores: The number of cores to use (if running locally), default is set to 90% of available.
    :return: A list of the 3D gas density profiles measured by this function, though if the measurement was not
        successful an entry of None will be added to the list.
    :rtype: List[GasDensity3D]
    """
    if annulus_method not in ALLOWED_ANN_METHODS:
        a_meth = ", ".join(ALLOWED_ANN_METHODS)
        raise ValueError("That is not a valid method for deciding where to place annuli, please use one of "
                         "these; {}".format(a_meth))

    if annulus_method == 'min_snr':
        # This returns the boundary radii for the annuli
        ann_rads = min_snr_proj_temp_prof(sources, outer_radii, min_snr, min_width, use_combined, use_worst, lo_en,
                                          hi_en, psf_corr, psf_model, psf_bins, psf_algo, psf_iter, allow_negative,
                                          exp_corr, group_spec, min_counts, min_sn, over_sample, one_rmf, freeze_met,
                                          abund_table, temp_lo_en, temp_hi_en, num_cores)
    elif annulus_method == 'min_cnt':
        # This returns the boundary radii for the annuli, based on a minimum number of counts per annulus
        ann_rads = min_cnt_proj_temp_prof(sources, outer_radii, min_cnt, min_width, use_combined, lo_en, hi_en,
                                          psf_corr, psf_model, psf_bins, psf_algo, psf_iter, group_spec, min_counts,
                                          min_sn, over_sample, one_rmf, freeze_met, abund_table, temp_lo_en, temp_hi_en,
                                          num_cores)
    elif annulus_method == "growth":
        raise NotImplementedError("This method isn't implemented yet")

    # So we can iterate through sources without worrying if there's more than one cluster
    if not isinstance(sources, ClusterSample):
        sources = [sources]

    # Don't need to check abundance table input because that happens in min_snr_proj_temp_prof and the
    #  gas_density_profile method of APECNormalisation1D
    final_dens_profs = []
    with tqdm(desc="Generating density profiles from annular spectra", total=len(sources)) as dens_prog:
        for src_ind, src in enumerate(sources):
            cur_rads = ann_rads[src_ind]

            try:
                # The normalisation profile(s) from the fit that produced the projected temperature profile.
                apec_norm_prof = src.get_apec_norm_profiles(cur_rads, group_spec, min_counts, min_sn, over_sample)

                obs_id = 'combined'
                inst = 'combined'
                # Seeing as we're here, I might as well make a density profile from the apec normalisation profile
                dens_prof = apec_norm_prof.gas_density_profile(src.redshift, src.cosmo, abund_table, num_data_real,
                                                               sigma, num_dens)
                # Then I store it in the source
                src.update_products(dens_prof)
                final_dens_profs.append(dens_prof)

            # It is possible that no normalisation profile exists because the spectral fitting failed, we account
            #  for that here
            except NoProductAvailableError:
                warn("{s} doesn't have a matching apec normalisation profile, skipping.")
                final_dens_profs.append(None)

            # It's also possible that the gas_density_profile method of our normalisation profile is going to
            #  throw a ValueError because some values are infinite or NaNs - we have to catch that too
            except ValueError:
                warn("{s}'s density profile has NaN values in it, skipping.", stacklevel=2)
                final_dens_profs.append(None)

            dens_prog.update(1)

    return final_dens_profs


def inv_abel_data(sources: Union[GalaxyCluster, ClusterSample], outer_radius: Union[str, Quantity],
                  inv_abel_method: str, num_dens: bool = True, use_peak: bool = True, pix_step: int = 1,
                  min_snr: Union[int, float] = 0.0, abund_table: str = "angr", lo_en: Quantity = Quantity(0.5, 'keV'),
                  hi_en: Quantity = Quantity(2.0, 'keV'), psf_corr: bool = True, psf_model: str = "ELLBETA",
                  psf_bins: int = 4, psf_algo: str = "rl", psf_iter: int = 15, num_samples: int = 10000,
                  group_spec: bool = True, min_counts: int = 5, min_sn: float = None, over_sample: float = None,
                  obs_id: Union[str, list] = None, inst: Union[str, list] = None, conv_temp: Quantity = None,
                  conv_outer_radius: Quantity = "r500", num_cores: int = NUM_CORES) -> List[GasDensity3D]:
    """
    A count-rate-map-based galaxy cluster gas density calculation method where a surface brightness profile inverse
    abel transformed, thus inferring the 3D count-rate/volume profile. Then a conversion factor calculated from
    simulated spectra is used to infer the number density profile.

    There are a number of choices of method for inverse abel transforming, provided by the Python package PyAbel:

    * direct - This attempts a direct integration of the inverse-Abel integral (see
      https://ned.ipac.caltech.edu/level5/March02/Sarazin/Sarazin5_5_4.html). No assumptions are made other than
      cylindrical symmetry, and fine sampling is required. This is the only method that supports non-uniform
      radius sampling, and if the surface brightness profile is detected to have non-uniform radius sampling (if
      generated for a minimum signal-to-noise for instance) then this is the method that will be used.

    * basex - This method (basis set expansion) uses functions with a known analytic inverse-abel
      transform (gaussian-like in this case).

    * hansen_law_ho0 - A fast transform method (the PyAbel authors describe it 'a hidden gem of the field', using
      a coordinate transformation to model the Abel transform as a set of linear differential equation. This
      particular choice uses hold_order=0, which assumes a constant intensity across a pixel (between grid points)
      for the driving function (the image gradient for the inverse transform).

    * hansen_law_ho1 - A fast transform method (the PyAbel authors describe it 'a hidden gem of the field', using
      a coordinate transformation to model the Abel transform as a set of linear differential equation. This
      particular choice uses hold_order=1, which assumes a linear intensity variation between grid points, which may
      yield a more accurate transform for some functions

    * onion_peeling - In the onion-peeling method the projection is approximated by rings of constant property,
      see the PyAbel documentation for the mathematical explanation.

    * two_point - The Abel integral is broken into intervals between the radius points, with the function
      assumed to be constant within the points.

    * three_point - This method exploits the observation that the value of the Abel inverted data at any radial
      position r is primarily determined from changes in the projection data in the neighborhood of r. The
      projection data (raw data) is expanded as a quadratic function of in the neighborhood of each data point,
      which allows an analytical deprojection.

    * daun - Similar to onion peeling, but uses 'Tikhonov regularization'.

    The PyAbel documentation provides a useful discussion of when and when not to use different methods
    (https://pyabel.readthedocs.io/en/latest/transform_methods.html), and the authors also wrote a paper comparing
    the various methods (https://arxiv.org/abs/1902.09007).

    :param GalaxyCluster/ClusterSample sources: A GalaxyCluster or ClusterSample object to measure density
        profiles for.
    :param str/Quantity outer_radius: The radius to which the surface brightness profile should be generated. It can
        be a named radius (e.g. 'r500', 'r200') or a quantity containing a value (or values, if a sample of clusters
        has been passed).
    :param str inv_abel_method: The method/algorithm which should be used for the inverse abel transform of the
        surface brightness profile data. We advise reading the information in the docstring before making a choice, as
        well as experimenting a little.
    :param bool num_dens: If True then a number density profile will be generated, otherwise a mass density profile
        will be generated.
    :param bool use_peak: If true the measured peak will be used as the central coordinate of the profile.
    :param int pix_step: The width (in pixels) of each annular bin for the profiles, default is 1.
    :param int/float min_snr: The minimum allowed signal-to-noise for the surface brightness
        profiles. Default is 0, which disables automatic re-binning.
    :param str abund_table: Which abundance table should be used for the XSPEC fit, FakeIt run, and for the
        electron/hydrogen number density ratio.
    :param Quantity lo_en: The lower energy limit of the combined ratemap used to calculate density.
    :param Quantity hi_en: The upper energy limit of the combined ratemap used to calculate density.
    :param bool psf_corr: Default True, whether PSF corrected ratemaps will be used to make the
        surface brightness profile, and thus the density (if False density results could be incorrect).
    :param str psf_model: If PSF corrected, the PSF model used.
    :param int psf_bins: If PSF corrected, the number of bins per side.
    :param str psf_algo: If PSF corrected, the algorithm used.
    :param int psf_iter: If PSF corrected, the number of algorithm iterations.
    :param int num_samples: The number of samples drawn from the posterior distributions of model parameters
        after the fitting process is complete.
    :param bool group_spec: Whether the spectra that were used for fakeit were grouped.
    :param float min_counts: The minimum counts per channel, if the spectra that were used for fakeit
        were grouped by minimum counts.
    :param float min_sn: The minimum signal-to-noise per channel, if the spectra that were used for fakeit
        were grouped by minimum signal-to-noise.
    :param float over_sample: The level of oversampling applied on the spectra that were used for fakeit.
    :param str/list obs_id: A specific ObsID(s) to measure the density from. This should be a string if a single
        source is being analysed, and a list of ObsIDs the same length as the number of sources otherwise. The
        default is None, in which case the combined data will be used to measure the density profile.
    :param str/list inst: A specific instrument(s) to measure the density from. This can either be passed as a
        single string (e.g. 'pn') if only one source is being analysed, or the same instrument should be used for
        every source in a sample, or a list of strings if different instruments are required for each source. The
        default is None, in which case the combined data will be used to measure the density profile.
    :param Quantity conv_temp: If set this will override XGA measured temperatures within the conv_outer_radius, and
        the fakeit run to calculate the normalisation conversion factor will use these temperatures. The quantity
         should have an entry for each cluster being analysed. Default is None.
    :param str/Quantity conv_outer_radius: The outer radius within which to generate spectra and measure temperatures
        for the conversion factor calculation, default is 'r500'. An astropy quantity may also be passed, with either
        a single value or an entry for each cluster being analysed.
    :param int num_cores: The number of cores that the evselect call and XSPEC functions are allowed to use.
    :return: A list of the 3D gas density profiles measured by this function, though if the measurement was not
        successful an entry of None will be added to the list.
    :rtype: List[GasDensity3D]
    """
    # Run the setup function, calculates the factors that translate 3D count-rate to density
    #  Also checks parameters and runs any spectra/fits that need running
    sources, conv_factors, obs_id, inst = _dens_setup(sources, outer_radius, Quantity(0, 'arcsec'), abund_table, lo_en,
                                                      hi_en, group_spec, min_counts, min_sn, over_sample, obs_id, inst,
                                                      conv_temp, conv_outer_radius, num_cores)

    # Calls the handy spectrum region setup function to make a predictable set of outer radius values
    out_rads = region_setup(sources, outer_radius, Quantity(0, 'arcsec'), False, '')[-1]

    # Before we properly get started, we just check that the user has passed an 'allowed' inverse-Abel transform
    if inv_abel_method not in ALLOWED_INV_ABEL:
        all_str = ", ".join(ALLOWED_INV_ABEL)
        raise ValueError("{p} is not a supported inverse-Abel transform method, please choose from; "
                         "{a}".format(p=inv_abel_method, a=all_str))

    with tqdm(desc="Inverse Abel transforming data and measuring densities",
              total=len(sources), position=0) as dens_prog:
        final_dens_profs = []
        # I need the ratio of electrons to protons here as well, so just fetch that for the current abundance table
        e_to_p_ratio = NHC[abund_table]
        for src_ind, src in enumerate(sources):

            sb_prof = _run_sb(src, out_rads[src_ind], use_peak, lo_en, hi_en, psf_corr, psf_model, psf_bins, psf_algo,
                              psf_iter, pix_step, min_snr, obs_id[src_ind], inst[src_ind])
            # It is possible for the above function to fail to produce an SB profile, in which case the return is
            #  None, so we check for that and don't continue to the rest of this function (for this source anyway) if
            #  it is None
            if sb_prof is None:
                final_dens_profs.append(None)
                continue
            else:
                # If it is a good SB profile, we make sure to add it to the source's storage structure
                src.update_products(sb_prof)

            # Sets up the resolution of the radial spatial sampling for the inverse-abel transform methods
            force_change = False
            if len(set(np.diff(sb_prof.radii.value).round(5))) != 1:
                warn("Most numerical methods for the abel transform require uniformly sampled radius values, setting "
                     "the method to 'direct'", stacklevel=2)
                inv_abel_method = 'direct'
                force_change = True
            else:
                dr = (sb_prof.radii[1] - sb_prof.radii[0]).value

            realisations = sb_prof.generate_data_realisations(num_samples)
            transform_res = np.zeros(realisations.shape)

            for t_ind in range(0, realisations.shape[0]):
                if inv_abel_method == 'direct' and force_change:
                    # This is necessary (see issue #1164) for the direct method because the last value is by definition
                    #  zero - one of the PyAbel authors suggested padding out the data.
                    to_trans = np.concatenate([realisations[t_ind, :], np.array([0.0])])
                    temp_dr = (sb_prof.radii[-1] - sb_prof.radii[-2]).value
                    mod_rad = np.concatenate([sb_prof.radii.value, np.array([sb_prof.radii.value[-1] + temp_dr])])
                    transform_res[t_ind, :] = direct_transform(to_trans, r=mod_rad, backend='python',
                                                               verbose=False)[:-1]
                elif inv_abel_method == 'direct' and not force_change:
                    # This is necessary (see issue #1164) for the direct method because the last value is by definition
                    #  zero - one of the PyAbel authors suggested padding out the data.
                    to_trans = np.concatenate([realisations[t_ind, :], np.array([0.0])])
                    transform_res[t_ind, :] = direct_transform(to_trans, dr=dr, verbose=False, backend='python')[:-1]
                elif inv_abel_method == 'basex':
                    transform_res[t_ind, :] = basex_transform(realisations[t_ind, :], dr=dr, verbose=False)
                elif inv_abel_method == 'hansen_law_ho0':
                    transform_res[t_ind, :] = hansenlaw_transform(realisations[t_ind, :], dr=dr, verbose=False)
                elif inv_abel_method == 'hansen_law_ho1':
                    transform_res[t_ind, :] = hansenlaw_transform(realisations[t_ind, :], dr=dr, hold_order=1,
                                                                  verbose=False)
                elif inv_abel_method == 'onion_bordas':
                    transform_res[t_ind, :] = onion_bordas_transform(realisations[t_ind, :], dr=dr, verbose=False)
                elif inv_abel_method == 'onion_peeling':
                    transform_res[t_ind, :] = onion_peeling_transform(realisations[t_ind, :], dr=dr, verbose=False)
                elif inv_abel_method == 'two_point':
                    transform_res[t_ind, :] = two_point_transform(realisations[t_ind, :], dr=dr, verbose=False)
                elif inv_abel_method == 'three_point':
                    transform_res[t_ind, :] = three_point_transform(realisations[t_ind, :], dr=dr, verbose=False)
                elif inv_abel_method == 'daun':
                    transform_res[t_ind, :] = daun_transform(realisations[t_ind, :], dr=dr, verbose=False)

            # The result is NO LONGER an astropy quantity, so we need to set that up again - we also transpose to
            #  orient it properly
            transformed = Quantity(transform_res, sb_prof.values_unit / sb_prof.radii_unit).T

            # Grab the radii that we need for the density profile we're about to set up
            dens_rads = sb_prof.radii.copy()
            dens_rads_errs = sb_prof.radii_err.copy()
            dens_deg_rads = sb_prof.deg_radii.copy()

            # Now need to make sure the units of the transformed model are what we need
            if sb_prof.values_unit.is_equivalent('ct/(s*arcmin**2)'):
                # If the SB profile is in count/s/arcmin^2 then the abel transform will have
                #  units of ct/s/(arcmin^2 kpc), so I create a quantity which will convert the arcmin^2 to kpc^2
                conv = Quantity(ang_to_rad(Quantity(1, 'arcmin'), src.redshift, src.cosmo).to("kpc").value,
                                'kpc/arcmin')**2
                transformed /= conv
            elif sb_prof.values_unit.is_equivalent('ct/(s*kpc**2)'):
                pass
            else:
                raise NotImplementedError("Haven't yet added support for surface brightness profiles in "
                                          "other units, don't really know how you even got here.")

            # We convert the volume element to cm^3 now, this is the unit we expect for the density conversion
            transformed = transformed.to('ct/(s*cm^3)')

            # We multiply by the conversion factor that is unique to the cluster and calculated earlier to take
            #  the transformed profile to a gas number density (n_gas as seen in Eckert et al. 2016, eq. 2).
            num_dens_dist = np.sqrt(transformed * conv_factors[src_ind])*(1+e_to_p_ratio)

            med_num_dens = np.nanpercentile(num_dens_dist, 50, axis=1)
            num_dens_err = np.nanstd(num_dens_dist, axis=1)

            # Setting up the instrument and ObsID to pass into the density profile definition
            if obs_id[src_ind] is None:
                cur_inst = "combined"
                cur_obs = "combined"
            else:
                cur_inst = inst[src_ind]
                cur_obs = obs_id[src_ind]

            try:
                # I now allow the user to decide if they want to generate number or mass density profiles using
                #  this function, and here is where that distinction is made
                if num_dens:
                    dens_prof = GasDensity3D(dens_rads.to("kpc"), med_num_dens, sb_prof.centre, src.name, cur_obs,
                                             cur_inst, inv_abel_method, sb_prof, dens_rads_errs, num_dens_err,
                                             deg_radii=dens_deg_rads, auto_save=True)
                else:
                    # TODO Check the origin of the mean molecular weight, see if there are different values for
                    #  different abundance tables
                    # The mean molecular weight multiplied by the proton mass
                    conv_mass = MEAN_MOL_WEIGHT*m_p
                    dens_prof = GasDensity3D(dens_rads.to("kpc"), (med_num_dens*conv_mass).to('Msun/Mpc^3'),
                                             sb_prof.centre, src.name, cur_obs, cur_inst, inv_abel_method, sb_prof,
                                             dens_rads_errs, (num_dens_err*conv_mass).to('Msun/Mpc^3'),
                                             deg_radii=dens_deg_rads, auto_save=True)

                src.update_products(dens_prof)
                final_dens_profs.append(dens_prof)

                # If, for some reason, there are some inf/NaN values in any of the quantities passed to the GasDensity3D
                #  declaration, this is where an error will be thrown
            except ValueError:
                final_dens_profs.append(None)
                warn("One or more of the quantities passed to the init of {}'s density profile has a NaN or Inf value"
                     " in it.".format(src.name), stacklevel=2)

            dens_prog.update(1)

    return final_dens_profs



#  This code is a part of X-ray: Generate and Analyse (XGA), a module designed for the XMM Cluster Survey (XCS).
#  Last modified by David J Turner (turne540@msu.edu) 25/07/2024, 14:27. Copyright (c) The Contributors
import inspect
from types import FunctionType
from typing import Tuple, Union
from warnings import warn

import numpy as np
import scipy.odr as odr
from astropy.units import Quantity, UnitConversionError
from scipy.optimize import curve_fit

from ..exceptions import XGAFunctionConversionError, XGAOptionalDependencyError
from ..models import convert_to_odr_compatible
from ..models.misc import power_law
from ..products.relation import ScalingRelation

# This is just so the allowed fitting methods can be referenced from one place
ALLOWED_FIT_METHODS = ["curve_fit", "odr", "lira", "emcee"]


def _fit_initialise(y_values: Quantity, y_errs: Quantity, x_values: Quantity, x_errs: Quantity = None,
                    y_norm: Quantity = None, x_norm: Quantity = None, log_data: bool = False,
                    point_names: Union[np.ndarray, list] = None, third_dim: Union[np.ndarray, Quantity, list] = None) \
        -> Tuple[Quantity, Quantity, Quantity, Quantity, Quantity, Quantity, np.ndarray, Quantity]:
    """
    A handy little function that prepares the data for fitting with the chosen method.

    :param Quantity y_values: The y data values to fit to.
    :param Quantity y_errs: The y error values of the data. These should be supplied as either a 1D Quantity with
        length N (where N is the length of y_values), or an Nx2 Quantity with lower and upper errors.
    :param Quantity x_values: The x data values to fit to.
    :param Quantity x_errs: The x error values of the data. These should be supplied as either a 1D Quantity with
        length N (where N is the length of x_values), or an Nx2 Quantity with lower and upper errors.
    :param Quantity y_norm: Quantity to normalise the y data by.
    :param Quantity x_norm: Quantity to normalise the x data by.
    :param bool log_data: This parameter controls whether the data is logged before being returned. The
        default is False as it isn't likely to be used often - its included because LIRA wants logged data.
    :param np.ndarray/list point_names: A possible set of source names associated with all the data points
    :param np.ndarray/Quantity/list third_dim: A possible set of extra data used to colour data points.
    :return: The x data, x errors, y data, and y errors. Also the x_norm, y_norm, and the names of non-NaN points.
    :rtype: Tuple[Quantity, Quantity, Quantity, Quantity, Quantity, Quantity, np.ndarray, Quantity]
    """
    # Check the lengths of the value and uncertainty quantities, as well as the extra information that can also
    #  flow through this function
    if len(x_values) != len(y_values):
        raise ValueError("The x and y quantities must have the same number of entries!")
    elif len(y_errs) != len(y_values):
        raise ValueError("Uncertainty quantities must have the same number of entries as the value quantities.")
    elif x_errs is not None and len(x_errs) != len(x_values):
        raise ValueError("Uncertainty quantities must have the same number of entries as the value quantities.")
    # Not involved in the fitting process, but comes through here so that the sources dropped due to NaN values
    #  also have the values dropped in these variables
    elif point_names is not None and len(point_names) != len(x_values):
        ValueError("The 'point_names' argument is a different length ({p}) to the input data "
                   "({d}).".format(p=len(point_names), d=len(x_values)))
    elif third_dim is not None and len(third_dim) != len(x_values):
        ValueError("The 'third_dim' argument is a different length ({p}) to the input data "
                   "({d}).".format(p=len(third_dim), d=len(x_values)))
    elif y_errs.unit != y_values.unit:
        raise UnitConversionError("Uncertainty quantities must have the same units as value quantities.")
    elif x_errs is not None and x_errs.unit != x_values.unit:
        raise UnitConversionError("Uncertainty quantities must have the same units as value quantities.")

    # For my own sanity, we're going to make x_err a Quantity if it isn't one already
    if x_errs is None:
        # We want zero error
        x_errs = Quantity(np.zeros(len(x_values)), x_values.unit)
        # We're also going to set a flag indicating that a default value was set
        no_x_errs = True
    else:
        no_x_errs = False

    # Need to do a cleaning stage, to remove any NaN values from the data
    #  First have to identify which entries in both the x and y arrays are NaN
    x_not_nans = np.where(~np.isnan(x_values))[0]
    y_not_nans = np.where(~np.isnan(y_values))[0]
    all_not_nans = np.intersect1d(x_not_nans, y_not_nans)

    # We also check for negative uncertainties, which are obviously bogus and cause plotting/fit issues later on - we
    #  take the opposite approach (in terms of boolean logic) to identifying the bad entries, because its easier
    x_err_are_neg = np.where(x_errs < 0)[0]
    y_err_are_neg = np.where(y_errs < 0)[0]

    all_err_not_neg = np.intersect1d(np.setdiff1d(np.arange(0, len(x_errs)), x_err_are_neg),
                                     np.setdiff1d(np.arange(0, len(y_errs)), y_err_are_neg))

    # Intersect the two selection criteria
    all_acc = np.intersect1d(all_not_nans, all_err_not_neg)

    # And we'll repeat the warning exercise if any were excluded because they have negative uncertainties
    thrown_away = len(x_values) - len(all_acc)
    if thrown_away != 0:
        warn("{} sources have NaN values or negative uncertainties and have been excluded".format(thrown_away),
             stacklevel=2)

    # Only values that aren't NaN and don't have negative errors will be permitted
    x_values = x_values[all_acc]
    y_values = y_values[all_acc]
    # We're not changing the error arrays here because I'll do that in the place where I ensure the error arrays
    #  are 1D

    # We need to see if the normalisation parameters have been set, and if not then make them a number
    #  And if they have been set we need to ensure they're the same units as the x and y values
    if x_norm is None:
        x_norm = Quantity(1, x_values.unit)
    elif x_norm is not None and x_norm.unit != x_values.unit:
        raise UnitConversionError("The x normalisation parameter must have the same units as the x values.")

    if y_norm is None:
        y_norm = Quantity(1, y_values.unit)
    elif y_norm is not None and y_norm.unit != y_values.unit:
        raise UnitConversionError("The y normalisation parameter must have the same units as the y values.")

    # We need to parse the errors that have been parsed in, because there are options here
    # The user can either pass a 1D quantity with the errors, in which case we assume that they are symmetrical
    if y_errs.ndim == 2:
        # Because people can't follow instructions, I'm going to try and automatically find which axis to
        #  average the errors over
        av_ax = y_errs.shape.index(2)
        y_errs = np.mean(y_errs, axis=av_ax)

    if x_errs.ndim == 2:
        # Because people can't follow instructions, I'm going to try and automatically find which axis to
        #  average the errors over
        av_ax = x_errs.shape.index(2)
        x_errs = np.mean(x_errs, axis=av_ax)

    # Doing what we did to the value arrays further up, removing any entries that haven't passed our criteria of not
    #  having a NaN and not having negative errors
    y_errs = y_errs[all_acc]
    x_errs = x_errs[all_acc]

    # We divide through by the normalisation parameter, which makes the data unitless
    x_fit_data = x_values / x_norm
    y_fit_data = y_values / y_norm
    x_fit_err = x_errs / x_norm
    y_fit_err = y_errs / y_norm

    # TODO I'M STILL NOT SURE THAT THIS IS THE RIGHT WAY TO CONVERT ERRORS TO LOG
    if log_data:
        # We're logging here because the fitting package wants it
        x_fit_err = x_fit_err / x_fit_data / np.log(10)
        x_fit_data = np.log10(x_fit_data)

        y_fit_err = y_fit_err / y_fit_data / np.log(10)
        y_fit_data = np.log10(y_fit_data)

        if no_x_errs:
            # I know I'm setting something that already exists, but I want errors of 0 to be passed out
            x_fit_err = Quantity(np.zeros(len(x_values)), x_values.unit)

    # Make sure point_names actually is an array (if supplied) and remove the NaN entry equivalents
    if point_names is not None:
        if isinstance(point_names, list):
            point_names = np.array(point_names)
        point_names = point_names[all_not_nans]
    elif point_names is None:
        point_names = None

    # Same deal with the third dimension data that can optionally be supplied to the scaling relations (though
    #  isn't used in the fit process, it's just for colouring data points in a view method).
    if third_dim is not None:
        if isinstance(third_dim, list):
            third_dim = Quantity(third_dim)
        third_dim = third_dim[all_not_nans]
    elif third_dim is None:
        third_dim = None

    return x_fit_data, x_fit_err, y_fit_data, y_fit_err, x_norm, y_norm, point_names, third_dim


def scaling_relation_curve_fit(model_func: FunctionType, y_values: Quantity, y_errs: Quantity, x_values: Quantity,
                               x_errs: Quantity = None, y_norm: Quantity = None, x_norm: Quantity = None,
                               x_lims: Quantity = None, start_pars: list = None, y_name: str = 'Y',
                               x_name: str = 'X', dim_hubb_ind: Union[float, int] = None,
                               point_names: Union[np.ndarray, list] = None,
                               third_dim_info: Union[np.ndarray, Quantity] = None, third_dim_name: str = None,
                               x_en_bounds: Quantity = None, y_en_bounds: Quantity = None) -> ScalingRelation:
    """
    A function to fit a scaling relation with the scipy non-linear least squares implementation (curve fit), generate
    an XGA ScalingRelation product, and return it.

    :param FunctionType model_func: The function object of the model you wish to fit. PLEASE NOTE, the function must
        be defined in the style used in xga.models.misc; i.e. powerlaw(x: np.ndarray, k: float, a: float), where
        the first argument is for x values, and the following arguments are all fit parameters.
    :param Quantity y_values: The y data values to fit to.
    :param Quantity y_errs: The y error values of the data. These should be supplied as either a 1D Quantity with
        length N (where N is the length of y_values), or an Nx2 Quantity with lower and upper errors.
    :param Quantity x_values: The x data values to fit to.
    :param Quantity x_errs: The x error values of the data. These should be supplied as either a 1D Quantity with
        length N (where N is the length of x_values), or an Nx2 Quantity with lower and upper errors.
    :param Quantity y_norm: Quantity to normalise the y data by.
    :param Quantity x_norm: Quantity to normalise the x data by.
    :param Quantity x_lims: The range of x values in which this relation is valid, default is None. If this
        information is supplied, please pass it as a Quantity array, with the first element being the lower
        bound and the second element being the upper bound.
    :param list start_pars: The start parameters for the curve_fit run, default is None, which means curve_fit
        will use all ones.
    :param str y_name: The name to be used for the y-axis of the scaling relation (DON'T include the unit, that
        will be inferred from the astropy Quantity.
    :param str x_name: The name to be used for the x-axis of the scaling relation (DON'T include the unit, that
        will be inferred from the astropy Quantity.
    :param float/int dim_hubb_ind: This is used to tell the ScalingRelation which power of E(z) has been applied
        to the y-axis data, this can then be used by the predict method to remove the E(z) contribution from
        predictions. The default is None.
    :param np.ndarray/list point_names: The source names associated with the data points passed in to this scaling
        relation, can be used for diagnostic purposes (i.e. identifying which source an outlier belongs to).
    :param np.ndarray/Quantity third_dim_info: A set of data points which represent a faux third dimension. They should
        not have been involved in the fitting process, and the relation should not be in three dimensions, but these
        can be used to colour the data points in a view method.
    :param str third_dim_name: The name of the third dimension data.
    :param Tuple[Quantity] x_en_bounds: If the value on the x-axis of this relation is 'energy bound', those bounds
        can be specified here (e.g. if the value is 0.5-2.0 keV luminosity you would pass a non-scalar quantity with
        the first entry being 0.5 and the second 2.0; Quantity([0.5, 2.0], 'keV'). The default is None.
    :param Tuple[Quantity] y_en_bounds: If the value on the y-axis of this relation is 'energy bound', those bounds
        can be specified here (e.g. if the value is 0.5-2.0 keV luminosity you would pass a non-scalar quantity with
        the first entry being 0.5 and the second 2.0; Quantity([0.5, 2.0], 'keV'). The default is None.
    :return: An XGA ScalingRelation object with all the information about the data and fit, a view method, and a
        predict method.
    :rtype: ScalingRelation
    """
    x_fit_data, x_fit_errs, y_fit_data, y_fit_errs, x_norm, y_norm, point_names, \
        third_dim_info = _fit_initialise(y_values, y_errs, x_values, x_errs, y_norm, x_norm, point_names=point_names,
                                         third_dim=third_dim_info)

    fit_par, fit_cov = curve_fit(model_func, x_fit_data.value, y_fit_data.value, sigma=y_fit_errs.value,
                                 absolute_sigma=True, p0=start_pars)
    fit_par_err = np.sqrt(np.diagonal(fit_cov))

    sr = ScalingRelation(fit_par, fit_par_err, model_func, x_norm, y_norm, x_name, y_name, fit_method='Curve Fit',
                         x_data=x_fit_data * x_norm, y_data=y_fit_data * y_norm, x_err=x_fit_errs * x_norm,
                         y_err=y_fit_errs * y_norm, x_lims=x_lims, dim_hubb_ind=dim_hubb_ind, point_names=point_names,
                         third_dim_info=third_dim_info, third_dim_name=third_dim_name, x_en_bounds=x_en_bounds,
                         y_en_bounds=y_en_bounds)

    return sr


def scaling_relation_odr(model_func: FunctionType, y_values: Quantity, y_errs: Quantity, x_values: Quantity,
                         x_errs: Quantity = None, y_norm: Quantity = None, x_norm: Quantity = None,
                         x_lims: Quantity = None, start_pars: list = None, y_name: str = 'Y',
                         x_name: str = 'X', dim_hubb_ind: Union[float, int] = None,
                         point_names: Union[np.ndarray, list] = None,
                         third_dim_info: Union[np.ndarray, Quantity] = None, third_dim_name: str = None,
                         x_en_bounds: Quantity = None, y_en_bounds: Quantity = None) -> ScalingRelation:
    """
    A function to fit a scaling relation with the scipy orthogonal distance regression implementation, generate
    an XGA ScalingRelation product, and return it.

    :param FunctionType model_func: The function object of the model you wish to fit. PLEASE NOTE, the function must
        be defined in the style used in xga.models.misc; i.e. powerlaw(x: np.ndarray, k: float, a: float), where
        the first argument is for x values, and the following arguments are all fit parameters. The scipy ODR
        implementation requires functions of a different style, and I try to automatically convert the input function
        to that style, but to help that please avoid using one letter parameter names in any custom function you might
        want to use.
    :param Quantity y_values: The y data values to fit to.
    :param Quantity y_errs: The y error values of the data. These should be supplied as either a 1D Quantity with
        length N (where N is the length of y_values), or an Nx2 Quantity with lower and upper errors.
    :param Quantity x_values: The x data values to fit to.
    :param Quantity x_errs: The x error values of the data. These should be supplied as either a 1D Quantity with
        length N (where N is the length of x_values), or an Nx2 Quantity with lower and upper errors.
    :param Quantity y_norm: Quantity to normalise the y data by.
    :param Quantity x_norm: Quantity to normalise the x data by.
    :param Quantity x_lims: The range of x values in which this relation is valid, default is None. If this
        information is supplied, please pass it as a Quantity array, with the first element being the lower
        bound and the second element being the upper bound.
    :param list start_pars: The start parameters for the ODR run, default is all ones.
    :param str y_name: The name to be used for the y-axis of the scaling relation (DON'T include the unit, that
        will be inferred from the astropy Quantity.
    :param str x_name: The name to be used for the x-axis of the scaling relation (DON'T include the unit, that
        will be inferred from the astropy Quantity.
    :param float/int dim_hubb_ind: This is used to tell the ScalingRelation which power of E(z) has been applied
        to the y-axis data, this can then be used by the predict method to remove the E(z) contribution from
        predictions. The default is None.
    :param np.ndarray/list point_names: The source names associated with the data points passed in to this scaling
        relation, can be used for diagnostic purposes (i.e. identifying which source an outlier belongs to).
    :param np.ndarray/Quantity third_dim_info: A set of data points which represent a faux third dimension. They should
        not have been involved in the fitting process, and the relation should not be in three dimensions, but these
        can be used to colour the data points in a view method.
    :param str third_dim_name: The name of the third dimension data.
    :param Tuple[Quantity] x_en_bounds: If the value on the x-axis of this relation is 'energy bound', those bounds
        can be specified here (e.g. if the value is 0.5-2.0 keV luminosity you would pass a non-scalar quantity with
        the first entry being 0.5 and the second 2.0; Quantity([0.5, 2.0], 'keV'). The default is None.
    :param Tuple[Quantity] y_en_bounds: If the value on the y-axis of this relation is 'energy bound', those bounds
        can be specified here (e.g. if the value is 0.5-2.0 keV luminosity you would pass a non-scalar quantity with
        the first entry being 0.5 and the second 2.0; Quantity([0.5, 2.0], 'keV'). The default is None.
    :return: An XGA ScalingRelation object with all the information about the data and fit, a view method, and a
        predict method.
    :rtype: ScalingRelation
    """
    if start_pars is None:
        # Setting up the start_pars, as ODR doesn't have a default value like curve_fit
        num_par = len(list(inspect.signature(model_func).parameters.keys())) - 1
        start_pars = np.ones(num_par)

    # Standard data preparation
    x_fit_data, x_fit_errs, y_fit_data, y_fit_errs, x_norm, y_norm, point_names, \
        third_dim_info = _fit_initialise(y_values, y_errs, x_values, x_errs, y_norm, x_norm, point_names=point_names,
                                         third_dim=third_dim_info)

    # Immediately faced with a problem, because scipy's ODR is naff and wants functions defined like
    #  blah(par_vector, x_values), which is completely different to my standard definition of models in this module.
    # I don't want the user to have to define things differently for this fit function, so I'm gonna try and
    #  dynamically redefine function models passed in here...
    converted_model_func = convert_to_odr_compatible(model_func)

    # The first thing I want to do is check that the newly converted function gives the same result for a
    #  simple test as the original model
    if model_func(5, *start_pars) != converted_model_func(start_pars, 5):
        raise XGAFunctionConversionError('I attempted to convert the input model function to the standard'
                                         ' required by ODR, but it is not returning the same value as the '
                                         'original for this test case.')

    # Then we define a scipy odr model
    odr_model = odr.Model(converted_model_func)

    # And a RealData (which takes the errors as the proper standard deviations of the data)
    odr_data = odr.RealData(x_fit_data, y_fit_data, x_fit_errs, y_fit_errs)

    # Now we instantiate the ODR class with our model and data. This is currently basically ripped from the example
    #  given in the scipy ODR documentation
    odr_obj = odr.ODR(odr_data, odr_model, beta0=start_pars)

    # Actually run the fit, and grab the output
    fit_results = odr_obj.run()

    # And from here, with this output object, I just read out the parameter values, along with the standard dev
    fit_par = fit_results.beta
    fit_par_err = fit_results.sd_beta

    sr = ScalingRelation(fit_par, fit_par_err, model_func, x_norm, y_norm, x_name, y_name, fit_method='ODR',
                         x_data=x_fit_data * x_norm, y_data=y_fit_data * y_norm, x_err=x_fit_errs * x_norm,
                         y_err=y_fit_errs * y_norm, x_lims=x_lims, odr_output=fit_results, dim_hubb_ind=dim_hubb_ind,
                         point_names=point_names, third_dim_info=third_dim_info, third_dim_name=third_dim_name,
                         x_en_bounds=x_en_bounds, y_en_bounds=y_en_bounds)

    return sr


def scaling_relation_lira(y_values: Quantity, y_errs: Quantity, x_values: Quantity, x_errs: Quantity = None,
                          y_norm: Quantity = None, x_norm: Quantity = None, x_lims: Quantity = None, y_name: str = 'Y',
                          x_name: str = 'X', num_steps: int = 100000, num_chains: int = 4, num_burn_in: int = 10000,
                          dim_hubb_ind: Union[float, int] = None, point_names: Union[np.ndarray, list] = None,
                          third_dim_info: Union[np.ndarray, Quantity] = None, third_dim_name: str = None,
                          x_en_bounds: Quantity = None, y_en_bounds: Quantity = None) -> ScalingRelation:
    """
    A function to fit a power law scaling relation with the excellent R fitting package LIRA
    (https://doi.org/10.1093/mnras/stv2374), this function requires a valid R installation, along with LIRA (and its
    dependencies such as JAGS), as well as the Python module rpy2.

    :param Quantity y_values: The y data values to fit to.
    :param Quantity y_errs: The y error values of the data. These should be supplied as either a 1D Quantity with
        length N (where N is the length of y_values), or an Nx2 Quantity with lower and upper errors.
    :param Quantity x_values: The x data values to fit to.
    :param Quantity x_errs: The x error values of the data. These should be supplied as either a 1D Quantity with
        length N (where N is the length of x_values), or an Nx2 Quantity with lower and upper errors.
    :param Quantity y_norm: Quantity to normalise the y data by.
    :param Quantity x_norm: Quantity to normalise the x data by.
    :param Quantity x_lims: The range of x values in which this relation is valid, default is None. If this
        information is supplied, please pass it as a Quantity array, with the first element being the lower
        bound and the second element being the upper bound.
    :param str y_name: The name to be used for the y-axis of the scaling relation (DON'T include the unit, that
        will be inferred from the astropy Quantity.
    :param str x_name: The name to be used for the x-axis of the scaling relation (DON'T include the unit, that
        will be inferred from the astropy Quantity.
    :param int num_steps: The number of steps to take in each chain.
    :param int num_chains: The number of chains to run.
    :param int num_burn_in: The number of steps to discard as a burn in period. This is also used as the adapt
        parameter of the LIRA fit.
    :param float/int dim_hubb_ind: This is used to tell the ScalingRelation which power of E(z) has been applied
        to the y-axis data, this can then be used by the predict method to remove the E(z) contribution from
        predictions. The default is None.
    :param np.ndarray/list point_names: The source names associated with the data points passed in to this scaling
        relation, can be used for diagnostic purposes (i.e. identifying which source an outlier belongs to).
    :param np.ndarray/Quantity third_dim_info: A set of data points which represent a faux third dimension. They should
        not have been involved in the fitting process, and the relation should not be in three dimensions, but these
        can be used to colour the data points in a view method.
    :param str third_dim_name: The name of the third dimension data.
    :param Tuple[Quantity] x_en_bounds: If the value on the x-axis of this relation is 'energy bound', those bounds
        can be specified here (e.g. if the value is 0.5-2.0 keV luminosity you would pass a non-scalar quantity with
        the first entry being 0.5 and the second 2.0; Quantity([0.5, 2.0], 'keV'). The default is None.
    :param Tuple[Quantity] y_en_bounds: If the value on the y-axis of this relation is 'energy bound', those bounds
        can be specified here (e.g. if the value is 0.5-2.0 keV luminosity you would pass a non-scalar quantity with
        the first entry being 0.5 and the second 2.0; Quantity([0.5, 2.0], 'keV'). The default is None.
    :return: An XGA ScalingRelation object with all the information about the data and fit, a view method, and a
        predict method.
    :rtype: ScalingRelation
    """
    # Due to the nature of this function, a wrapper for the LIRA R fitting package, I'm going to try the
    #  necessary imports here, because the external dependencies are likely to only be used in this function
    #  and as such I don't want to force the user to have them to use XGA.
    try:
        from rpy2.robjects.packages import importr
        from rpy2 import robjects
        robjects.r['options'](warn=-1)
    except ImportError:
        raise XGAOptionalDependencyError('LIRA is an R fitting package, and as such you need to have installed '
                                         'rpy2 to use this function')

    # We use the rpy2 module to interface with an underlying R installation, and import the basic R components
    base_pack = importr('base')
    utils_pack = importr('utils')

    # Now we import the thing we're actually interested in, the LIRA package
    try:
        lira_pack = importr('lira')
    except robjects.packages.PackageNotInstalledError:
        raise XGAOptionalDependencyError('While the rpy2 module is installed, you do not appear to have installed '
                                         'the LIRA fitting package to your R environment')

    # Slightly different data preparation to the other fitting methods, this one returns logged data and errors
    x_fit_data, x_fit_errs, y_fit_data, y_fit_errs, x_norm, y_norm, point_names, \
        third_dim_info = _fit_initialise(y_values, y_errs, x_values, x_errs, y_norm, x_norm, True, point_names,
                                         third_dim_info)

    # And now we have to make some R objects so that we can pass it through our R interface to the LIRA package
    x_fit_data = robjects.FloatVector(x_fit_data.value)
    y_fit_data = robjects.FloatVector(y_fit_data.value)
    x_fit_errs = robjects.FloatVector(x_fit_errs.value)
    y_fit_errs = robjects.FloatVector(y_fit_errs.value)

    # This runs the LIRA fit and grabs the output data frame, from that I can read the chains for the different
    #  parameters
    chains = lira_pack.lira(x_fit_data, y_fit_data, delta_x=x_fit_errs, delta_y=y_fit_errs, n_iter=num_steps,
                            n_chains=num_chains, n_adapt=num_burn_in, export=False, print_summary=False,
                            print_diagnostic=False)[0][0]

    # Read out the alpha parameter chain and convert to a numpy array
    alpha_par_chain = np.power(10, np.array(chains.rx2['alpha.YIZ']))
    alpha_par_val = np.mean(alpha_par_chain)
    alpha_par_err = np.std(alpha_par_chain)

    # Read out the beta parameter chain and convert to a numpy array
    beta_par_chain = np.array(chains.rx2['beta.YIZ'])
    beta_par_val = np.mean(beta_par_chain)
    beta_par_err = np.std(beta_par_chain)

    # Read out the intrinsic scatter chain and convert to a numpy array
    sigma_par_chain = np.array(chains.rx2['sigma.YIZ.0'])
    sigma_par_val = np.mean(sigma_par_chain)
    sigma_par_err = np.std(sigma_par_chain)

    fit_par = np.array([beta_par_val, alpha_par_val])
    fit_par_err = np.array([beta_par_err, alpha_par_err])

    # This call to the fit initialisation function DOESN'T produce logged data, do this so the plot works
    #  properly - it expects non logged data
    x_fit_data, x_fit_errs, y_fit_data, y_fit_errs, x_norm, y_norm, \
        throw_away, sec_throw_away = _fit_initialise(y_values, y_errs, x_values, x_errs, y_norm, x_norm)

    # I'm re-formatting the chains into a shape that the ScalingRelation class will understand.
    xga_chains = np.concatenate([beta_par_chain.reshape(len(beta_par_chain), 1),
                                 alpha_par_chain.reshape(len(alpha_par_chain), 1)], axis=1)

    sr = ScalingRelation(fit_par, fit_par_err, power_law, x_norm, y_norm, x_name, y_name, fit_method='LIRA',
                         x_data=x_fit_data * x_norm, y_data=y_fit_data * y_norm, x_err=x_fit_errs * x_norm,
                         y_err=y_fit_errs * y_norm, x_lims=x_lims, chains=xga_chains,
                         scatter_par=np.array([sigma_par_val, sigma_par_err]), scatter_chain=sigma_par_chain,
                         dim_hubb_ind=dim_hubb_ind, point_names=point_names, third_dim_info=third_dim_info,
                         third_dim_name=third_dim_name, x_en_bounds=x_en_bounds, y_en_bounds=y_en_bounds)

    return sr


def scaling_relation_emcee():
    raise NotImplementedError("This fitting method has not yet been implemented, please consider LIRA "
                              "as an alternative.")



#  This code is a part of X-ray: Generate and Analyse (XGA), a module designed for the XMM Cluster Survey (XCS).
#  Last modified by David J Turner (turne540@msu.edu) 22/04/2023, 01:09. Copyright (c) The Contributors

from .lightcurve import *
from ._common import region_setup
from .phot import *
from .spec import *



#  This code is a part of X-ray: Generate and Analyse (XGA), a module designed for the XMM Cluster Survey (XCS).
#  Last modified by David J Turner (turne540@msu.edu) 01/08/2024, 14:13. Copyright (c) The Contributors

from typing import Tuple, List, Union
from warnings import warn, simplefilter

import numpy as np
from astropy import wcs
from astropy.coordinates import SkyCoord
from astropy.cosmology import Cosmology
from astropy.units import Quantity, UnitBase, deg, UnitConversionError
from numpy import ndarray

from .base import BaseSource
from .. import DEFAULT_COSMO
from ..exceptions import NotAssociatedError, PeakConvergenceFailedError, NoRegionsError, NoValidObservationsError, \
    NoProductAvailableError
from ..products import RateMap
from ..sourcetools import rad_to_ang, ang_to_rad, nh_lookup

# This disables an annoying astropy warning that pops up all the time with XMM images
# Don't know if I should do this really
simplefilter('ignore', wcs.FITSFixedWarning)


class ExtendedSource(BaseSource):
    """
    The general extended source XGA class, for extended X-ray sources that you do not yet want to analyse as
    a specific type of astrophysical object. This class is subclassed by GalaxyCluster, which then adds more
    specific analyses, for instance. This general class is useful for when you're doing exploratory analyses.

    :param float ra: The right-ascension of the source, in degrees.
    :param float dec: The declination of the source, in degrees.
    :param float redshift: The redshift of the source, optional. Default is None.
    :param str name: The name of the source, optional. Name will be constructed from position if None.
    :param Quantity custom_region_radius: A custom analysis region radius for this source, optional.
    :param bool use_peak: Whether peak position should be found and used.
    :param Quantity peak_lo_en: The lower energy bound for the RateMap to calculate peak position
        from. Default is 0.5keV
    :param Quantity peak_hi_en: The upper energy bound for the RateMap to calculate peak position
        from. Default is 2.0keV.
    :param float back_inn_rad_factor: This factor is multiplied by an analysis region radius, and gives the inner
        radius for the background region. Default is 1.05.
    :param float back_out_rad_factor: This factor is multiplied by an analysis region radius, and gives the outer
        radius for the background region. Default is 1.5.
    :param cosmology: An astropy cosmology object for use throughout analysis of the source.
    :param bool load_products: Whether existing products should be loaded from disk.
    :param bool load_fits: Whether existing fits should be loaded from disk.
    :param str peak_find_method: Which peak finding method should be used (if use_peak is True). Default
        is hierarchical, simple may also be passed.
    :param bool in_sample: A boolean argument that tells the source whether it is part of a sample or not, setting
        to True suppresses some warnings so that they can be displayed at the end of the sample progress bar. Default
        is False. User should only set to True to remove warnings.
    """
    def __init__(self, ra: float, dec: float, redshift: float = None, name: str = None,
                 custom_region_radius: Quantity = None, use_peak: bool = True,
                 peak_lo_en: Quantity = Quantity(0.5, "keV"), peak_hi_en: Quantity = Quantity(2.0, "keV"),
                 back_inn_rad_factor: float = 1.05, back_out_rad_factor: float = 1.5,
                 cosmology: Cosmology = DEFAULT_COSMO,  load_products: bool = True, load_fits: bool = False,
                 peak_find_method: str = "hierarchical", in_sample: bool = False):
        """
        The init for the general extended source XGA class, takes information on the position (and optionally
        redshift) of source of interest, matches to extended regions, and optionally performs peak finding.

        :param float ra: The right-ascension of the source, in degrees.
        :param float dec: The declination of the source, in degrees.
        :param float redshift: The redshift of the source, optional. Default is None.
        :param str name: The name of the source, optional. Name will be constructed from position if None.
        :param Quantity custom_region_radius: A custom analysis region radius for this source, optional.
        :param bool use_peak: Whether peak position should be found and used.
        :param Quantity peak_lo_en: The lower energy bound for the RateMap to calculate peak position
            from. Default is 0.5keV
        :param Quantity peak_hi_en: The upper energy bound for the RateMap to calculate peak position
            from. Default is 2.0keV.
        :param float back_inn_rad_factor: This factor is multiplied by an analysis region radius, and gives the inner
            radius for the background region. Default is 1.05.
        :param float back_out_rad_factor: This factor is multiplied by an analysis region radius, and gives the outer
            radius for the background region. Default is 1.5.
        :param Cosmology cosmology: An astropy cosmology object for use throughout analysis of the source.
        :param bool load_products: Whether existing products should be loaded from disk.
        :param bool load_fits: Whether existing fits should be loaded from disk.
        :param str peak_find_method: Which peak finding method should be used (if use_peak is True). Default
            is hierarchical, simple may also be passed.
        :param bool in_sample: A boolean argument that tells the source whether it is part of a sample or not, setting
            to True suppresses some warnings so that they can be displayed at the end of the sample progress bar. Default
            is False. User should only set to True to remove warnings.
        """
        # Calling the BaseSource init method
        super().__init__(ra, dec, redshift, name, cosmology, load_products, load_fits, in_sample)

        self._custom_region_radius = None
        # Setting up the custom region radius attributes
        if custom_region_radius is not None and custom_region_radius.unit.is_equivalent("kpc"):
            rad = rad_to_ang(custom_region_radius, self._redshift, self._cosmo).to("deg")
            self._custom_region_radius = rad
            self._radii["custom"] = self._custom_region_radius
            self._rad_info = True
        elif custom_region_radius is not None and not custom_region_radius.unit.is_equivalent("kpc"):
            self._custom_region_radius = custom_region_radius.to("deg")
            self._radii["custom"] = self._custom_region_radius
            self._rad_info = True

        # Adding a custom radius to act as a search aperture for peak finding
        # 500kpc in degrees, for the current redshift and cosmology
        #  Or 5 arcminutes if no redshift information is present (that is allowed for the ExtendedSource class)
        if self._redshift is not None:
            search_aperture = rad_to_ang(Quantity(500, "kpc"), self._redshift, cosmo=self._cosmo)
        else:
            search_aperture = Quantity(5, 'arcmin').to('deg')
        self._radii["search"] = search_aperture

        self._use_peak = use_peak
        self._back_inn_factor = back_inn_rad_factor
        self._back_out_factor = back_out_rad_factor
        # Make sure the peak energy boundaries are in keV
        self._peak_lo_en = peak_lo_en.to('keV')
        self._peak_hi_en = peak_hi_en.to('keV')
        self._peaks = {o: {} for o in self.obs_ids}
        self._peaks.update({"combined": None})
        self._peaks_near_edge = {o: {} for o in self.obs_ids}
        self._peaks_near_edge.update({"combined": None})
        self._chosen_peak_cluster = None
        self._other_peak_clusters = None
        self._snr = {}

        # This uses the added context of the type of source to find (or not find) matches in region files
        self._regions, self._alt_match_regions, self._other_regions = self._source_type_match("ext")

        # Making a combined list of interloper regions
        self._interloper_regions = []
        for o in self._other_regions:
            self._interloper_regions += self._other_regions[o]

        # Run through any alternative matches and raise warnings
        for o in self._alt_match_regions:
            if len(self._alt_match_regions[o]) > 0:
                warn_text = "There are {0} alternative matches for observation {1}, associated with " \
                            "source {2}".format(len(self._alt_match_regions[o]), o, self.name)
                if not self._samp_member:
                    warn(warn_text, stacklevel=2)
                else:
                    self._supp_warn.append(warn_text)

        if not in_sample:
            # We make sure there are some images in existence, using the peak finding energy bounds
            from xga.sas import evselect_image
            evselect_image(self, self.peak_lo_en, self.peak_hi_en)

        self._interloper_masks = {}
        for obs_id in self.obs_ids:
            # Generating and storing these because they should only
            cur_im = self.get_products("image", obs_id)[0]
            self._interloper_masks[obs_id] = self._generate_interloper_mask(cur_im)

        # Constructs the detected dictionary, detailing whether the source has been detected IN REGION FILES
        #  in each observation.
        self._detected = {o: self._regions[o] is not None for o in self._regions}

        # If in some of the observations the source has not been detected, a warning will be raised
        if True in self._detected.values() and False in self._detected.values():
            warn_text = "{n} has not been detected in all region files, so generating and fitting products" \
                        " with the 'region' reg_type will not use all available data".format(n=self.name)
            if not self._samp_member:
                warn(warn_text, stacklevel=2)
            else:
                self._supp_warn.append(warn_text)

        # If the source wasn't detected in ALL of the observations, then we have to rely on a custom region,
        #  and if no custom region options are passed by the user then an error is raised.
        elif all([det is False for det in self._detected.values()]) and self._custom_region_radius is not None:
            warn_text = "{n} has not been detected in ANY region files, so generating and fitting products" \
                        " with the 'region' reg_type will not work".format(n=self.name)
            if not self._samp_member:
                warn(warn_text, stacklevel=2)
            else:
                self._supp_warn.append(warn_text)
        elif all([det is False for det in self._detected.values()]) and self._custom_region_radius is None \
                and "GalaxyCluster" not in repr(self):
            raise NoRegionsError("{n} has not been detected in ANY region files, and no custom region or "
                                 " overdensity radius has been passed. No analysis is possible.".format(n=self.name))

        # Call to a method that goes through all the observations and finds the X-ray peak. Also at the same
        #  time finds the X-ray peak of the combined ratemap (an essential piece of information).
        self._all_peaks(peak_find_method)
        if self._use_peak:
            self._default_coord = self.peak

    def find_peak(self, rt: RateMap, method: str = "hierarchical", num_iter: int = 20, peak_unit: UnitBase = deg) \
            -> Tuple[Quantity, bool, bool, ndarray, List]:
        """
        A method that will find the X-ray peak for the RateMap that has been passed in. It takes
        the user supplied coordinates from source initialisation as a starting point, finds the peak within a 500kpc
        radius, re-centres the region, and iterates until the peak converges to within 15kpc, or until 20
        20 iterations has been reached.

        :param RateMap rt: The ratemap which we want to find the peak (local to our user supplied coordinates) of.
        :param str method: Which peak finding method to use. Currently either hierarchical or simple can be chosen.
        :param int num_iter: How many iterations should be allowed before the peak is declared as not converged.
        :param UnitBase peak_unit: The unit the peak coordinate is returned in.
        :return: The peak coordinate, a boolean flag as to whether the returned coordinates are near
            a chip gap/edge, and a boolean flag as to whether the peak converged. It also returns the coordinates
            of the points within the chosen point cluster, and a list of all point clusters that were not chosen.
        :rtype: Tuple[Quantity, bool, bool, ndarray, List]
        """
        all_meth = ["hierarchical", "simple"]
        if method not in all_meth:
            raise ValueError("{0} is not a recognised, use one of the following methods: "
                             "{1}".format(method, ", ".join(all_meth)))
        central_coords = SkyCoord(*self.ra_dec.to("deg"))

        # Iteration counter just to kill it if it doesn't converge
        count = 0
        # Allow 20 iterations by default before we kill this - alternatively loop will exit when centre converges
        #  to within 15kpc (or 0.15arcmin).
        while count < num_iter:
            aperture_mask = self.get_mask("search", rt.obs_id, central_coords)[0]

            # Find the peak using the experimental clustering_peak method
            if method == "hierarchical":
                try:
                    peak, near_edge, chosen_coords, other_coords = rt.clustering_peak(aperture_mask, peak_unit)
                except ValueError:
                    raise PeakConvergenceFailedError("The hierarchical clustering peak finder does not "
                                                     "have enough points to work with.")
            elif method == "simple":
                peak, near_edge = rt.simple_peak(aperture_mask, peak_unit)
                chosen_coords = []
                other_coords = []

            peak_deg = rt.coord_conv(peak, deg)
            # Calculate the distance between new peak and old central coordinates
            separation = Quantity(np.sqrt(abs(peak_deg[0].value - central_coords.ra.value) ** 2 +
                                          abs(peak_deg[1].value - central_coords.dec.value) ** 2), deg)

            central_coords = SkyCoord(*peak_deg.copy())
            if self._redshift is not None:
                separation = ang_to_rad(separation, self._redshift, self._cosmo)

            if count != 0 and self._redshift is not None and separation <= Quantity(15, "kpc"):
                break
            elif count != 0 and self._redshift is None and separation <= Quantity(0.15, 'arcmin'):
                break

            count += 1

        if count == num_iter:
            converged = False
            # To do the least amount of damage, if the peak doesn't converge then we just return the
            #  user supplied coordinates
            peak = self.ra_dec
            near_edge = rt.near_edge(peak)
        else:
            converged = True

        return peak, near_edge, converged, chosen_coords, other_coords

    def _all_peaks(self, method: str):
        """
        An internal method that finds the X-ray peaks for all of the available observations and instruments,
        as well as the combined ratemap. Peak positions for individual ratemap products are allowed to not
        converge, and will just write None to the peak dictionary, but if the peak of the combined ratemap fails
        to converge an error will be thrown. The combined ratemap peak will also be stored by itself in an
        attribute, to allow a property getter easy access.

        :param str method: The method to be used for peak finding.
        """
        en_key = "bound_{l}-{u}".format(l=self._peak_lo_en.value, u=self._peak_hi_en.value)
        comb_rt = [rt[-1] for rt in self.get_products("combined_ratemap", just_obj=False) if en_key in rt]

        if len(comb_rt) != 0:
            comb_rt = comb_rt[0]
        else:
            # I didn't want to import this here, but otherwise circular imports become a problem
            from xga.sas import emosaic
            emosaic(self, "image", self._peak_lo_en, self._peak_hi_en, disable_progress=True)
            emosaic(self, "expmap", self._peak_lo_en, self._peak_hi_en, disable_progress=True)
            comb_rt = [rt[-1] for rt in self.get_products("combined_ratemap", just_obj=False) if en_key in rt][0]

        if self._use_peak:
            coord, near_edge, converged, cluster_coords, other_coords = self.find_peak(comb_rt, method=method)
            # Updating nH for new coord, probably won't make a difference most of the time
            self._nH = nh_lookup(coord)[0]
        else:
            # If we don't care about peak finding then this is the one to go for
            coord = self.ra_dec
            near_edge = comb_rt.near_edge(coord)
            converged = True
            cluster_coords = np.ndarray([])
            other_coords = []

        # Unfortunately if the peak convergence fails for the combined ratemap I have to raise an error
        if converged:
            self._peaks["combined"] = coord
            self._peaks_near_edge["combined"] = near_edge
            # I'm only going to save the point cluster positions for the combined ratemap
            self._chosen_peak_cluster = cluster_coords
            self._other_peak_clusters = other_coords
        else:
            raise PeakConvergenceFailedError("Peak finding on the combined ratemap failed to converge within "
                                             "15kpc for {n} in the {l}-{u} energy "
                                             "band.".format(n=self.name, l=self._peak_lo_en, u=self._peak_hi_en))

        # for obs in self.obs_ids:
        #     for rt in self.get_products("ratemap", obs_id=obs, extra_key=en_key, just_obj=True):
        #         if self._use_peak:
        #             coord, near_edge, converged, cluster_coords, other_coords = self.find_peak(rt)
        #             if converged:
        #                 self._peaks[obs][rt.instrument] = coord
        #                 self._peaks_near_edge[obs][rt.instrument] = near_edge
        #             else:
        #                 self._peaks[obs][rt.instrument] = None
        #                 self._peaks_near_edge[obs][rt.instrument] = None
        #         else:
        #             self._peaks[obs][rt.instrument] = self.ra_dec
        #             self._peaks_near_edge[obs][rt.instrument] = rt.near_edge(self.ra_dec)

    def get_peaks(self, obs_id: str = None, inst: str = None) -> Quantity:
        """
        A get method to return the peak of the X-ray emission of this GalaxyCluster.

        :param str obs_id: The ObsID to return the X-ray peak coordinates for.
        :param str inst: The instrument to return the X-ray peak coordinates for.
        :return: The X-ray peak coordinates for the input parameters.
        :rtype: Quantity
        """
        # Common sense checks, are the obsids/instruments associated with this source etc.
        if obs_id is not None and obs_id not in self.obs_ids:
            raise NotAssociatedError("The ObsID {o} is not associated with {s}.".format(o=obs_id, s=self.name))
        elif obs_id is None and inst is not None:
            raise ValueError("If obs_id is None, inst cannot be None as well.")
        elif obs_id is not None and inst is not None and inst not in self._peaks[obs_id]:
            raise NotAssociatedError("The instrument {i} is not associated with observation {o} of this "
                                     "source.".format(i=inst, o=obs_id))
        elif obs_id is None and inst is None:
            chosen = self._peaks
        elif obs_id is not None and inst is None:
            chosen = self._peaks[obs_id]
        else:
            chosen = self._peaks[obs_id][inst]

        return chosen

    def get_1d_brightness_profile(self, outer_rad: Union[Quantity, str], obs_id: str = 'combined',
                                  inst: str = 'combined', central_coord: Quantity = None, radii: Quantity = None,
                                  lo_en: Quantity = None, hi_en: Quantity = None, pix_step: int = 1,
                                  min_snr: Union[float, int] = 0.0, psf_corr: bool = False, psf_model: str = "ELLBETA",
                                  psf_bins: int = 4, psf_algo: str = "rl", psf_iter: int = 15):
        """
        A specific get method for 1D brightness profiles. Should provide a relatively simple way of retrieving
        specific brightness profiles from XGA's storage system. Please note that there is not a separate get method
        for brightness profiles made from combined data, instead this method will search for combined profiles if
        either obs_id or inst is set to 'combined'. - Retrieving combined profiles is the default behaviour.

        :param Quantity/str outer_rad: The outermost radius of the profile, either as a Quantity or a name (e.g. r500).
        :param str obs_id: The ObsID used to generate the profile in question, default is None. If this is set to
            combined then this method will search for profiles based on combined data.
        :param str inst: The instrument used to generate the profile in question, default is None. If this is set to
            combined then this method will search for profiles based on combined data.
        :param Quantity central_coord: The central coordinate from which the profile was generated. Default is
            None, which means we shall use the default coordinate of this source.
        :param Quantity radii: Specific radii to check for in the profiles.
        :param Quantity lo_en: The lower energy bound of the RateMap used to generate the profile.
        :param Quantity hi_en: The upper energy bound of the RateMap used to generate the profile.
        :param int pix_step: The width of each annulus in pixels used to generate the profile.
        :param float min_snr: The minimum signal to noise imposed upon the profile.
        :param bool psf_corr: Is the brightness profile corrected for PSF effects?
        :param str psf_model: If PSF corrected, the PSF model used.
        :param int psf_bins: If PSF corrected, the number of bins per side.
        :param str psf_algo: If PSF corrected, the algorithm used.
        :param int psf_iter: If PSF corrected, the number of algorithm iterations.
        :return:
        """
        # Makes sure its in our standard unit
        if isinstance(outer_rad, str):
            outer_rad = self.get_radius(outer_rad, 'deg')
        elif isinstance(outer_rad, Quantity):
            outer_rad = self.convert_radius(outer_rad, 'deg')
        else:
            raise ValueError("Outer radius may only be a string or an astropy quantity")

        if obs_id == "combined" or inst == "combined":
            interim_prods = self.get_combined_profiles("brightness", central_coord, radii, lo_en, hi_en)
        else:
            interim_prods = self.get_profiles("brightness", obs_id, inst, central_coord, radii, lo_en, hi_en)

        # The methods I used to get this far will already have gotten upset if there are no matches, so I don't need
        #  to check they exist, but I do need to check if I have a list or a single object
        if not isinstance(interim_prods, list):
            interim_prods = [interim_prods]

        matched_prods = []
        for p in interim_prods:
            if not psf_corr and p.outer_radius == outer_rad and p.pix_step == pix_step and p.min_snr == min_snr \
                    and p.psf_corrected == psf_corr:
                matched_prods.append(p)
            elif psf_corr and p.outer_radius == outer_rad and p.pix_step == pix_step and p.min_snr == min_snr \
                    and p.psf_corrected == psf_corr and p.psf_model == psf_model and p.psf_bins == psf_bins \
                    and p.psf_algorithm == psf_algo and p.psf_iterations == psf_iter:
                matched_prods.append(p)

        if len(matched_prods) == 1:
            matched_prods = matched_prods[0]
        elif len(matched_prods) == 0:
            raise NoProductAvailableError("Cannot find any brightness profiles matching your input.")

        return matched_prods

    # Property SPECIFICALLY FOR THE COMBINED PEAK - as this is the peak we should be using mostly.
    @property
    def peak(self) -> Quantity:
        """
        A property getter for the combined X-ray peak coordinates. Most analysis will be centered
        on these coordinates.

        :return: The X-ray peak coordinates for the combined ratemap.
        :rtype: Quantity
        """
        return self._peaks["combined"]

    # I'm allowing this a setter, as some users may want to update the peak from outside (as is
    #  done in the ClusterSample init).
    @peak.setter
    def peak(self, new_peak: Quantity):
        """
        Allows the user to update the peak value used during analyses manually.

        :param Quantity new_peak: A new RA-DEC peak coordinate, in degrees.
        """
        if not new_peak.unit.is_equivalent("deg"):
            raise UnitConversionError("The new peak value must be in RA and DEC coordinates")
        elif len(new_peak) != 2:
            raise ValueError("Please pass an astropy Quantity, in units of degrees, with two entries - "
                             "one for RA and one for DEC.")
        self._peaks["combined"] = new_peak.to("deg")

    @property
    def custom_radius(self) -> Quantity:
        """
        A getter for the custom region that can be defined on initialisation.

        :return: The radius (in kpc) of the user defined custom region.
        :rtype: Quantity
        """
        return self._custom_region_radius

    @property
    def point_clusters(self) -> Tuple[ndarray, List[ndarray]]:
        """
        This allows you to retrieve the point cluster positions from the hierarchical clustering
        peak finding method run on the combined ratemap. This includes both the chosen cluster and
        all others that were found.

        :return: A numpy array of the positions of points of the chosen cluster (not galaxy cluster,
            a cluster of points). A list of numpy arrays with the same information for all the other clusters
            that were found
        :rtype: Tuple[ndarray, List[ndarray]]
        """
        return self._chosen_peak_cluster, self._other_peak_clusters


class PointSource(BaseSource):
    """
    The general point source XGA class, for point X-ray sources that you do not yet want to analyse as
    a specific type of astrophysical object. This general class is useful for when you're doing exploratory analyses.

    :param float ra: The right-ascension of the point source, in degrees.
    :param float dec: The declination of the point source, in degrees.
    :param float redshift: The redshift of the point source, optional. Default is None.
    :param str name: The name of the point source, optional. If no names are supplied
        then they will be constructed from the supplied coordinates.
    :param Quantity point_radius: The point source analysis region radius for this sample. An astropy quantity
        containing the radius should be passed; remember that units like kpc will also need redshift
        information. Default is 30 arcsecond radius.
    :param bool use_peak: Whether peak position should be found and used. For PointSource the 'simple' peak
        finding method is the only one available, other methods are allowed for extended sources.
    :param Quantity peak_lo_en: The lower energy bound for the RateMap to calculate peak
        position from. Default is 0.5keV.
    :param Quantity peak_hi_en: The upper energy bound for the RateMap to calculate peak
        position from. Default is 2.0keV.
    :param float back_inn_rad_factor: This factor is multiplied by an analysis region radius, and gives the inner
        radius for the background region. Default is 1.05.
    :param float back_out_rad_factor: This factor is multiplied by an analysis region radius, and gives the outer
        radius for the background region. Default is 1.5.
    :param Cosmology cosmology: An astropy cosmology object for use throughout analysis of the source.
    :param bool load_products: Whether existing products should be loaded from disk.
    :param bool load_fits: Whether existing fits should be loaded from disk.
    :param bool regen_merged: Should merged images/exposure maps be regenerated after cleaning. Default is
        True. This option is here so that sample objects can regenerate all merged products at once, which is
        more efficient as it can exploit parallelisation more fully - user probably doesn't need to touch this.
    :param bool in_sample: A boolean argument that tells the source whether it is part of a sample or not, setting
        to True suppresses some warnings so that they can be displayed at the end of the sample progress bar. Default
        is False. User should only set to True to remove warnings.
    """
    def __init__(self, ra, dec, redshift=None, name=None, point_radius=Quantity(30, 'arcsec'), use_peak=False,
                 peak_lo_en=Quantity(0.5, "keV"), peak_hi_en=Quantity(2.0, "keV"), back_inn_rad_factor=1.05,
                 back_out_rad_factor=1.5, cosmology: Cosmology = DEFAULT_COSMO, load_products=True, load_fits=False,
                 regen_merged: bool = True, in_sample: bool = False):
        """
        The init of the general XGA point source class.

        :param float ra: The right-ascension of the point source, in degrees.
        :param float dec: The declination of the point source, in degrees.
        :param float redshift: The redshift of the point source, optional. Default is None.
        :param str name: The name of the point source, optional. If no names are supplied
            then they will be constructed from the supplied coordinates.
        :param Quantity point_radius: The point source analysis region radius for this sample. An astropy quantity
            containing the radius should be passed; remember that units like kpc will also need redshift
            information. Default is 30 arcsecond radius.
        :param bool use_peak: Whether peak position should be found and used. For PointSource the 'simple' peak
            finding method is the only one available, other methods are allowed for extended sources.
        :param Quantity peak_lo_en: The lower energy bound for the RateMap to calculate peak
            position from. Default is 0.5keV.
        :param Quantity peak_hi_en: The upper energy bound for the RateMap to calculate peak
            position from. Default is 2.0keV.
        :param float back_inn_rad_factor: This factor is multiplied by an analysis region radius, and gives the inner
            radius for the background region. Default is 1.05.
        :param float back_out_rad_factor: This factor is multiplied by an analysis region radius, and gives the outer
            radius for the background region. Default is 1.5.
        :param cosmology: An astropy cosmology object for use throughout analysis of the source.
        :param bool load_products: Whether existing products should be loaded from disk.
        :param bool load_fits: Whether existing fits should be loaded from disk.
        :param bool regen_merged: Should merged images/exposure maps be regenerated after cleaning. Default is
            True. This option is here so that sample objects can regenerate all merged products at once, which is
            more efficient as it can exploit parallelisation more fully - user probably doesn't need to touch this.
        :param bool in_sample: A boolean argument that tells the source whether it is part of a sample or not, setting
            to True suppresses some warnings so that they can be displayed at the end of the sample progress bar. Default
            is False. User should only set to True to remove warnings.
        """

        super().__init__(ra, dec, redshift, name, cosmology, load_products, load_fits, in_sample)
        # This uses the added context of the type of source to find (or not find) matches in region files
        # This is the internal dictionary where all regions, defined by reg-files or by users, will be stored
        self._regions, self._alt_match_regions, self._other_regions = self._source_type_match("pnt")
        self._detected = {o: self._regions[o] is not None for o in self._regions}

        # Making a combined list of interloper regions
        self._interloper_regions = []
        for o in self._other_regions:
            self._interloper_regions += self._other_regions[o]

        if point_radius is not None and point_radius.unit.is_equivalent("kpc"):
            rad = rad_to_ang(point_radius, self._redshift, self._cosmo).to("deg")
            self._custom_region_radius = rad
            self._radii["point"] = self._custom_region_radius
            self._rad_info = True
        elif point_radius is not None and not point_radius.unit.is_equivalent("kpc"):
            self._custom_region_radius = point_radius.to("deg")
            self._radii["point"] = self._custom_region_radius
            self._rad_info = True

        if self._redshift is not None and point_radius.unit.is_equivalent("kpc"):
            search_aperture = rad_to_ang(point_radius.to("kpc"), self._redshift, cosmo=self._cosmo)
        elif point_radius.unit.is_equivalent("deg"):
            search_aperture = point_radius.to("deg")
        else:
            raise UnitConversionError("Can't convert {u} to a XGA supported length unit".format(u=point_radius.unit))
        self._radii["search"] = search_aperture

        if not in_sample:
            # We make sure there are some images in existence, using the peak finding energy bounds
            from xga.sas import evselect_image
            evselect_image(self, self.peak_lo_en, self.peak_hi_en)

        # This generates masks to remove interloper regions
        self._interloper_masks = {}
        for obs_id in self.obs_ids:
            # Generating and storing these because they should only
            cur_im = self.get_products("image", obs_id)[0]
            self._interloper_masks[obs_id] = self._generate_interloper_mask(cur_im)

        # Here we automatically clean the observations, to make sure the point source does actually lie
        #  on the detector and not just near it
        # Use a pretty harsh acceptance fraction
        reject_dict = self.obs_check("point", 0.9)
        if len(reject_dict) != 0:
            # Use the source method to remove data we've decided isn't worth keeping
            self.disassociate_obs(reject_dict)
            if len(self._obs) == 0:
                raise NoValidObservationsError("Observation cleaning has been run and there are no remaining"
                                               " observations. ")

            if regen_merged:
                from ..sas import emosaic
                emosaic(self, "image", self._peak_lo_en, self._peak_hi_en, disable_progress=True)
                emosaic(self, "expmap", self._peak_lo_en, self._peak_hi_en, disable_progress=True)

        self._use_peak = use_peak
        self._back_inn_factor = back_inn_rad_factor
        self._back_out_factor = back_out_rad_factor
        # Make sure the peak energy boundaries are in keV
        self._peak_lo_en = peak_lo_en.to('keV')
        self._peak_hi_en = peak_hi_en.to('keV')
        self._peaks = {o: {} for o in self.obs_ids}
        self._peaks.update({"combined": None})
        self._peaks_near_edge = {o: {} for o in self.obs_ids}

        self._all_peaks()

        if self._use_peak:
            self._default_coord = self.peak

    @property
    def point_radius(self) -> Quantity:
        """
        Property getter to access the point_radius declared on initialisation of the source, the radius
        of the region that is used for point source analysis.

        :return: The radius of the point source analysis region.
        :rtype: Quantity
        """
        return self._custom_region_radius

    def _all_peaks(self):
        en_key = "bound_{l}-{u}".format(l=self._peak_lo_en.value, u=self._peak_hi_en.value)
        comb_rt = self.get_products("combined_ratemap", extra_key=en_key)

        if len(comb_rt) != 0:
            comb_rt = comb_rt[0]
        else:
            from xga.sas import emosaic
            emosaic(self, "image", self._peak_lo_en, self._peak_hi_en, disable_progress=True)
            emosaic(self, "expmap", self._peak_lo_en, self._peak_hi_en, disable_progress=True)
            comb_rt = self.get_products("combined_ratemap", extra_key=en_key)[0]

        if self._use_peak:
            coord, near_edge = self.find_peak(comb_rt)
            # Updating nH for new coord, probably won't make a difference most of the time
            self._nH = nh_lookup(coord)[0]
        else:
            # If we don't care about peak finding then this is the boi to go for
            coord = self.ra_dec
            near_edge = comb_rt.near_edge(coord)

        self._peaks["combined"] = coord
        self._peaks_near_edge["combined"] = near_edge

    def find_peak(self, rt: RateMap, peak_unit: UnitBase = deg) -> Tuple[Quantity, bool]:
        """
        Uses a simple 'brightest pixel' method to measure a peak coordinate for the point source.

        :param RateMap rt: The RateMap to measure the peak from.
        :param UnitBase peak_unit: The desired output unit of the peak.
        :return:  The peak, and a boolean flag as to whether the peak is near an edge.
        :rtype: Tuple[Quantity, bool]
        """
        central_coords = SkyCoord(*self.ra_dec.to("deg"))
        aperture_mask = self.get_mask("search", rt.obs_id, central_coords)[0]
        peak, near_edge = rt.simple_peak(aperture_mask, peak_unit)

        return peak, near_edge

    @property
    def peak(self) -> Quantity:
        """
        A property getter for the combined X-ray peak coordinates.

        :return: The X-ray peak coordinates for the combined ratemap.
        :rtype: Quantity
        """
        return self._peaks["combined"]










#  This code is a part of X-ray: Generate and Analyse (XGA), a module designed for the XMM Cluster Survey (XCS).
#  Last modified by David J Turner (turne540@msu.edu) 20/11/2024, 22:32. Copyright (c) The Contributors

from typing import Union, List

import numpy as np
from astropy.units import Quantity, Unit, UnitConversionError, kpc, deg

from .base import BaseModel1D
from ..utils import r500, r200, r2500


class NFW(BaseModel1D):
    """
    A simple model to fit galaxy cluster mass profiles (https://ui.adsabs.harvard.edu/abs/1997ApJ...490..493N/abstract)
    - a cumulative mass version of the Navarro-Frenk-White profile. Typically, the NFW is formulated in terms of mass
    density, but one can derive a mass profile from it (https://ui.adsabs.harvard.edu/abs/2006MNRAS.368..518V/abstract).

    The NFW is extremely widely used, though generally for dark matter mass profiles, but will act as a handy
    functional form to fit to data-driven mass profiles derived from X-ray observations of clusters.

    :param Unit/str x_unit: The unit of the x-axis of this model, kpc for instance. May be passed as a string
        representation or an astropy unit object.
    :param Unit/str y_unit: The unit of the output of this model, Msun for instance. May be passed as a string
        representation or an astropy unit object.
    :param List[Quantity] cust_start_pars: The start values of the model parameters for any fitting function that
        used start values. The units are checked against default start values.
    """
    def __init__(self, x_unit: Union[str, Unit] = 'kpc', y_unit: Union[str, Unit] = Unit('Msun'),
                 cust_start_pars: List[Quantity] = None):
        """
        The init of a subclass of the XGA BaseModel1D class, describing the shape of cumulative mass profiles for
        a galaxy cluster based on the NFW mass density profile.
        """

        # If a string representation of a unit was passed then we make it an astropy unit
        if isinstance(x_unit, str):
            x_unit = Unit(x_unit)
        if isinstance(y_unit, str):
            y_unit = Unit(y_unit)

        poss_y_units = [Unit('Msun')]
        y_convertible = [u.is_equivalent(y_unit) for u in poss_y_units]
        if not any(y_convertible):
            allowed = ", ".join([u.to_string() for u in poss_y_units])
            raise UnitConversionError("{p} is not convertible to any of the allowed units; "
                                      "{a}".format(p=y_unit.to_string(), a=allowed))
        else:
            yu_ind = y_convertible.index(True)

        poss_x_units = [kpc, deg, r200, r500, r2500]
        x_convertible = [u.is_equivalent(x_unit) for u in poss_x_units]
        if not any(x_convertible):
            allowed = ", ".join([u.to_string() for u in poss_x_units])
            raise UnitConversionError("{p} is not convertible to any of the allowed units; "
                                      "{a}".format(p=x_unit.to_string(), a=allowed))
        else:
            xu_ind = x_convertible.index(True)

        r_scale_starts = [Quantity(100, 'kpc'), Quantity(0.2, 'deg'), Quantity(0.05, r200), Quantity(0.1, r500),
                         Quantity(0.5, r2500)]
        # We will implement the NFW mass profile with a rho_0 normalization parameter, a density - and leave in the
        #  volume integration terms - rather than fitting for some mass normalization
        norm_starts = [Quantity(1e+13, 'Msun/Mpc^3')]

        start_pars = [r_scale_starts[xu_ind], norm_starts[yu_ind]]
        if cust_start_pars is not None:
            # If the custom start parameters can run this gauntlet without tripping an error then we're all good
            # This method also returns the custom start pars converted to exactly the same units as the default
            start_pars = self.compare_units(cust_start_pars, start_pars)

        r_core_priors = [{'prior': Quantity([0, 2000], 'kpc'), 'type': 'uniform'},
                         {'prior': Quantity([0, 1], 'deg'), 'type': 'uniform'},
                         {'prior': Quantity([0, 1], r200), 'type': 'uniform'},
                         {'prior': Quantity([0, 1], r500), 'type': 'uniform'},
                         {'prior': Quantity([0, 1], r2500), 'type': 'uniform'}]
        norm_priors = [{'prior': Quantity([1e+12, 1e+16], 'Msun/Mpc^3'), 'type': 'uniform'}]

        priors = [r_core_priors[xu_ind], norm_priors[yu_ind]]

        nice_pars = [r"R$_{\rm{s}}$", r"$\rho_{0}$"]
        info_dict = {'author': 'Navarro J, Frenk C, White S', 'year': '1997',
                     'reference': 'https://ui.adsabs.harvard.edu/abs/1997ApJ...490..493N/abstract',
                     'general': 'The cumulative mass version of the NFW mass-density profile for galaxy \n'
                                'clusters - normally used to describe dark matter profiles.'}

        super().__init__(x_unit, y_unit, start_pars, priors, 'nfw', 'NFW Profile', nice_pars, 'Mass',
                         info_dict)

    @staticmethod
    def model(x: Quantity, r_scale: Quantity, rho_zero: Quantity) -> Quantity:
        """
        The model function for the constant-core and power-law entropy model.

        :param Quantity x: The radii to calculate y values for.
        :param Quantity r_scale: The scale radius parameter.
        :param Quantity rho_zero: A density normalization parameter.
        :return: The y values corresponding to the input x values.
        :rtype: Quantity
        """

        norm_rad = x / r_scale
        result = 4*np.pi*rho_zero*np.power(r_scale, 3)*(np.log(1 + norm_rad) - (norm_rad / (1 + norm_rad)))
        return result


# So that things like fitting functions can be written generally to support different models
MASS_MODELS = {"nfw": NFW}
MASS_MODELS_PUB_NAMES = {n: m().publication_name for n, m in MASS_MODELS.items()}
MASS_MODELS_PAR_NAMES = {n: m().par_publication_names for n, m in MASS_MODELS.items()}


#  This code is a part of X-ray: Generate and Analyse (XGA), a module designed for the XMM Cluster Survey (XCS).
#  Last modified by David J Turner (turne540@msu.edu) 07/11/2023, 10:57. Copyright (c) The Contributors


class HeasoftError(Exception):
    def __init__(self, *args):
        """
        Exception raised for unexpected output from HEASOFT calls, currently all encompassing.

        :param expression:
        :param message:
        """
        if args:
            self.message = args[0]
        else:
            self.message = None

    def __str__(self):
        if self.message:
            return '{0} '.format(self.message)
        else:
            return 'HeasoftError has been raised'


class SASNotFoundError(Exception):
    def __init__(self, *args):
        """
        Exception raised if the XMM Scientific Analysis System can not be found on the system.

        :param expression:
        :param message:
        """
        if args:
            self.message = args[0]
        else:
            self.message = None

    def __str__(self):
        if self.message:
            return '{0} '.format(self.message)
        else:
            return 'SASNotFoundError has been raised'


class XSPECNotFoundError(Exception):
    def __init__(self, *args):
        """
        Exception raised if XSPEC can not be found on the system.

        :param expression:
        :param message:
        """
        if args:
            self.message = args[0]
        else:
            self.message = None

    def __str__(self):
        if self.message:
            return '{0} '.format(self.message)
        else:
            return 'XSPECNotFoundError has been raised'


# I do not know if I will keep this as is or expand out into different errors
# The trouble is there are many hundreds of possible SAS errors, and I don't know if I
# want a class for all of them
class SASGenerationError(Exception):
    def __init__(self, *args):
        """
        Exception raised if an error is found to have occured during a run of a part of the SAS software.

        :param expression:
        :param message:
        """
        if args:
            self.message = args[0]
        else:
            self.message = None

    def __str__(self):
        if self.message:
            return '{0} '.format(self.message)
        else:
            return 'A generic SASNotFoundError has been raised'


class UnknownCommandlineError(Exception):
    def __init__(self, *args):
        """
        Exception raised if an error is found to have occured during a run of a part
        of the SAS software, but it cannot be linked to a SAS function.

        :param expression:
        :param message:
        """
        if args:
            self.message = args[0]
        else:
            self.message = None

    def __str__(self):
        if self.message:
            return '{0} '.format(self.message)
        else:
            return 'A generic UnknownCommandlineError has been raised'


class FailedProductError(Exception):
    def __init__(self, *args):
        """
        Exception raised when trying to access certain data/attributes from an object wrapping a
         product that failed to generate properly.

        :param expression:
        :param message:
        """
        if args:
            self.message = args[0]
        else:
            self.message = None

    def __str__(self):
        if self.message:
            return '{0} '.format(self.message)
        else:
            return 'FailedProductError has been raised.'


class XGAConfigError(Exception):
    def __init__(self, *args):
        """
        Exception raised for flawed XGA config files.

        :param expression:
        :param message:
        """
        if args:
            self.message = args[0]
        else:
            self.message = None

    def __str__(self):
        if self.message:
            return '{0} '.format(self.message)
        else:
            return 'XGAConfig has been raised'


class NoMatchFoundError(Exception):
    def __init__(self, *args):
        """
        Exception raised when source ra and dec coordinates can't be made to match to any XMM observation.

        :param expression:
        :param message:
        """
        if args:
            self.message = args[0]
        else:
            self.message = None

    def __str__(self):
        if self.message:
            return '{0} '.format(self.message)
        else:
            return 'NoMatchFoundError has been raised'


class NotAssociatedError(Exception):
    def __init__(self, *args):
        """
        Error raised when a given ObsID is not associated with a source object.

        :param expression:
        :param message:
        """
        if args:
            self.message = args[0]
        else:
            self.message = None

    def __str__(self):
        if self.message:
            return '{0} '.format(self.message)
        else:
            return 'NotAssociatedError has been raised'


class UnknownProductError(Exception):
    def __init__(self, *args):
        """
        Error raised when there is an attempt to write an unknown XMM product type to an XGA source.

        :param expression:
        :param message:
        """
        if args:
            self.message = args[0]
        else:
            self.message = None

    def __str__(self):
        if self.message:
            return '{}'.format(self.message)
        else:
            return 'UnknownProductTypeError has been raised'


class NoValidObservationsError(Exception):
    def __init__(self, *args):
        """
        Error raised when there is an initial match to an XMM observation, but the
        necessary files cannot be found.

        :param expression:
        :param message:
        """
        if args:
            self.message = args[0]
        else:
            self.message = None

    def __str__(self):
        if self.message:
            return '{}'.format(self.message)
        else:
            return 'NoValidObservationsError has been raised'


class MultipleMatchError(Exception):
    def __init__(self, *args):
        """
        Error raised when more than one match for a specific type of object is found in a single region file.
        Hopefully this error is never raised, it probably shouldn't!

        :param expression:
        :param message:
        """
        if args:
            self.message = args[0]
        else:
            self.message = None

    def __str__(self):
        if self.message:
            return '{}'.format(self.message)
        else:
            return 'MultipleMatchError has been raised'


class NoRegionsError(Exception):
    def __init__(self, *args):
        """
        Error raised when there are no appropriate regions available for an attempted analysis.

        :param expression:
        :param message:
        """
        if args:
            self.message = args[0]
        else:
            self.message = None

    def __str__(self):
        if self.message:
            return '{}'.format(self.message)
        else:
            return 'NoRegionsError has been raised'


class NoProductAvailableError(Exception):
    def __init__(self, *args):
        """
        Error raised when requesting a product from an XGA source that hasn't yet been generated.

        :param expression:
        :param message:
        """
        if args:
            self.message = args[0]
        else:
            self.message = None

    def __str__(self):
        if self.message:
            return '{}'.format(self.message)
        else:
            return 'NoProductAvailableError has been raised'


class ModelNotAssociatedError(Exception):
    def __init__(self, *args):
        """
        Error raised when values from a model fit that isn't associated with a particular product object.

        :param expression:
        :param message:
        """
        if args:
            self.message = args[0]
        else:
            self.message = None

    def __str__(self):
        if self.message:
            return '{}'.format(self.message)
        else:
            return 'ModelNotAssociatedError has been raised'


class ParameterNotAssociatedError(Exception):
    def __init__(self, *args):
        """
        Error raised when a parameter is not associated with a particular model.

        :param expression:
        :param message:
        """
        if args:
            self.message = args[0]
        else:
            self.message = None

    def __str__(self):
        if self.message:
            return '{}'.format(self.message)
        else:
            return 'ParameterNotAssociatedError has been raised'


class XSPECFitError(Exception):
    def __init__(self, *args):
        """
        This error is raised when there is a problem during an XSPEC fit instigated by XGA.

        :param expression:
        :param message:
        """
        if args:
            self.message = args[0]
        else:
            self.message = None

    def __str__(self):
        if self.message:
            return '{}'.format(self.message)
        else:
            return 'XSPECFitError has been raised'


class RateMapPairError(Exception):
    def __init__(self, *args):
        """
        This error is raised when there is a problem with the pair of Image and Exposure map objects that
        are passed into the RateMap init.

        :param expression:
        :param message:
        """
        if args:
            self.message = args[0]
        else:
            self.message = None

    def __str__(self):
        if self.message:
            return '{}'.format(self.message)
        else:
            return 'RateMapPairError has been raised'


class PeakConvergenceFailedError(Exception):
    def __init__(self, *args):
        """
        This error is raised when iterating peak finding fails to converge within the allowed number
        of iterations.

        :param expression:
        :param message:
        """
        if args:
            self.message = args[0]
        else:
            self.message = None

    def __str__(self):
        if self.message:
            return '{}'.format(self.message)
        else:
            return 'PeakConvergenceFailedError has been raised'


class SASInputInvalid(Exception):
    def __init__(self, *args):
        """
        This error is raised when a user provides an invalid input to a SAS function.

        :param expression:
        :param message:
        """
        if args:
            self.message = args[0]
        else:
            self.message = None

    def __str__(self):
        if self.message:
            return '{}'.format(self.message)
        else:
            return 'SASInputInvalid has been raised'


class NotPSFCorrectedError(Exception):
    def __init__(self, *args):
        """
        Raised when the user tries to set PSF deconvolution properties of an Image product, but the
        psf correction flag indicates that the product is not deconvolved.

        :param expression:
        :param message:
        """
        if args:
            self.message = args[0]
        else:
            self.message = None

    def __str__(self):
        if self.message:
            return '{}'.format(self.message)
        else:
            return 'NotPSFCorrectedError has been raised'


class IncompatibleProductError(Exception):
    def __init__(self, *args):
        """
        Raised when products are used together that do not have matching ObsID and instrument
        values, or when they were not generated at the same coordinates. For instance when you try to
        re-sample a PSF with an Image object from a different observation and instrument.

        :param expression:
        :param message:
        """
        if args:
            self.message = args[0]
        else:
            self.message = None

    def __str__(self):
        if self.message:
            return '{}'.format(self.message)
        else:
            return 'IncompatibleProductError has been raised'


class XGAFitError(Exception):
    def __init__(self, *args):
        """
        Raised when there is an issue with a fit that XGA is trying to perform.

        :param expression:
        :param message:
        """
        if args:
            self.message = args[0]
        else:
            self.message = None

    def __str__(self):
        if self.message:
            return '{}'.format(self.message)
        else:
            return 'XGAFitError has been raised'


class XGAInvalidModelError(Exception):
    def __init__(self, *args):
        """
        Raised when the user is trying to fit a model that is not appropriate to the data.

        :param expression:
        :param message:
        """
        if args:
            self.message = args[0]
        else:
            self.message = None

    def __str__(self):
        if self.message:
            return '{}'.format(self.message)
        else:
            return 'XGAInvalidModelError has been raised'


class XGAFunctionConversionError(Exception):
    def __init__(self, *args):
        """
        Raised when an attempt to convert an XGA model function (or a custom function supplied by the user) to the
        standard required by scipy's orthogonal distance regression failed.

        :param expression:
        :param message:
        """
        if args:
            self.message = args[0]
        else:
            self.message = None

    def __str__(self):
        if self.message:
            return '{}'.format(self.message)
        else:
            return 'XGAFunctionConversionError has been raised'


class XGAOptionalDependencyError(Exception):
    def __init__(self, *args):
        """
        Raised when an optional XGA dependency hasn't been installed, but the feature that needs it has been
        activated.

        :param expression:
        :param message:
        """
        if args:
            self.message = args[0]
        else:
            self.message = None

    def __str__(self):
        if self.message:
            return '{}'.format(self.message)
        else:
            return 'XGAOptionalDependencyError has been raised'


class XGASetIDError(Exception):
    def __init__(self, *args):
        """
        Raised when something is attempting to use annular spectra (which are all part of spectrum sets) together,
        but they have different set IDs.
        :param expression:
        :param message:
        """
        if args:
            self.message = args[0]
        else:
            self.message = None

    def __str__(self):
        if self.message:
            return '{}'.format(self.message)
        else:
            return 'XGASetIDError has been raised'


class XGAPoorDataError(Exception):
    def __init__(self, *args):
        """
        Raised when the data aren't of sufficient quality to complete an operation.
        :param expression:
        :param message:
        """
        if args:
            self.message = args[0]
        else:
            self.message = None

    def __str__(self):
        if self.message:
            return '{}'.format(self.message)
        else:
            return 'XGAPoorDataError has been raised'


class InvalidProductError(Exception):
    def __init__(self, *args):
        """
        Raised when there is a signficiant problem with a data product.
        :param expression:
        :param message:
        """
        if args:
            self.message = args[0]
        else:
            self.message = None

    def __str__(self):
        if self.message:
            return '{}'.format(self.message)
        else:
            return 'InvalidProductError has been raised'


class NotSampleMemberError(Exception):
    def __init__(self, *args):
        """
        Raised when a feature reserved for sources that belong to samples is used, and the source is independent
        of a sample.
        :param expression:
        :param message:
        """
        if args:
            self.message = args[0]
        else:
            self.message = None

    def __str__(self):
        if self.message:
            return '{}'.format(self.message)
        else:
            return 'NotSampleMemberError has been raised'




#  This code is a part of X-ray: Generate and Analyse (XGA), a module designed for the XMM Cluster Survey (XCS).
#  Last modified by David J Turner (turne540@msu.edu) 23/07/2024, 12:47. Copyright (c) The Contributors
from . import _version
__version__ = _version.get_versions()['version']

from .utils import xga_conf, CENSUS, OUTPUT, NUM_CORES, XGA_EXTRACT, BASE_XSPEC_SCRIPT, MODEL_PARS, \
    MODEL_UNITS, ABUND_TABLES, XSPEC_FIT_METHOD, COUNTRATE_CONV_SCRIPT, NHC, BLACKLIST, HY_MASS, MEAN_MOL_WEIGHT, \
    SAS_VERSION, XSPEC_VERSION, SAS_AVAIL, DEFAULT_COSMO
